# -*- coding: utf-8 -*-
# encoding: utf-8
"""
Plotting functions for pymzML.
The Factory object can hold several plots with several data traces each.
The data can be visualized as an interactive plotly plot or be exported as JSON.
"""

# Python mzML module - pymzml
# Copyright (C) 2010-2019 M. Kösters, C. Fufezan
#     The MIT License (MIT)

#     Permission is hereby granted, free of charge, to any person obtaining a copy
#     of this software and associated documentation files (the "Software"), to deal
#     in the Software without restriction, including without limitation the rights
#     to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#     copies of the Software, and to permit persons to whom the Software is
#     furnished to do so, subject to the following conditions:

#     The above copyright notice and this permission notice shall be included in all
#     copies or substantial portions of the Software.

#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#     IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#     FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#     AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#     LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#     OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#     SOFTWARE.

import sys
import math
import warnings


# Fail gracefully if no plotly installed
try:
    import plotly as plt

    # import plotly.offline as plt
    import plotly.graph_objs as go
    from plotly import subplots
except ImportError:
    warnings.warn("Plotly is required for plotting support.", ImportWarning)


from . import spec


class Factory(object):
    def __init__(self, filename=None):
        """
        Interface to visualize m/z or profile data using plotly (https://plot.ly/).

        Arguments:
            filename (str): Name for the output file. Default = "spectrum_plot.html"

        """
        self.filename = filename
        self.plots = []
        self.titles = []
        self.lookup = dict()
        self.y_max = []
        self.y_min = []
        self.x_max = []
        self.x_min = []
        self.offset = 1
        self.MS_precisions = []
        self.function_mapper = {
            "-__splineOffset__0": self.__return_neg_offset_0,
            "max_intensity": self.__return_max_y,
        }
        self.style_options = {"line.width": 1}  # default value

    def __return_max_y(self, i):
        """ """
        return self.y_max[i]

    def __return_neg_offset_0(self, i):
        """ """
        return 0.0 - (self.y_max[i] * (self.offset * 0.05))

    def new_plot(self, MS_precision="5e-6", title=None):
        """
        Add new plot to the plotting Factory.
        Every plot will be put into the x * 2 grid of one single figure.

        Keyword Arguments:
            title (str): an optional title that will be printed above the plot
            MS_precision (float): measuring MS_precision used in handler.
                Default 5e-6.

        Note:
            Old function newPlot() is still working. However, the new syntax
            should be used.

        """
        self.MS_precisions.append(MS_precision)
        self.plots.append([])
        self.titles.append(title)
        return

    def newPlot(self, MS_precision="5e-6", title=None):
        """Deprecated since version 1.2."""
        # Is this ok, pass the Factory.self to Spectrum class???
        # Id just make deprecation warning a function independent of any class
        # Also, this just looks wrong
        spec.Spectrum.deprecation_warning(
            self, function_name=sys._getframe().f_code.co_name
        )
        self.new_plot(MS_precision=MS_precision, title=title)

    def add(
        self,
        data,
        color=(0, 0, 0),
        style="sticks",
        mz_range=None,
        int_range=None,
        opacity=0.8,
        dash="solid",
        name=None,
        plot_num=-1,
        title=None,
    ):
        """
        Add data to the graph.

        Arguments:
            data (list): The data added to the graph. Must be list of  tuples with
                the following format. Note that i can be set to 'max_intensity' for
                setting the label position to the maximum intensity.
                    *   (mz,i) for all styles, except label,
                    *   (mz,i, string) for label.hoverinfo, .sticks and .triangle
                    *   (mz1, mz2, i, string) for label.linear and .spline

        Keyword Arguments:
            color (tuple): color encoded in RGB. Default = (0,0,0)
            style (str): plotting style. Default = "sticks".
                Currently supported styles are:
                    *   'lines':
                        Datapoints connected by lines
                    *   'points':
                        Datapoints without connection
                    *   'sticks':
                        Vertical line at given m/z (corresponds to centroided peaks)
                    *   'triangle' (MS_precision, micro, tiny, small, medium, big):
                        The top of the triangle corresponds to the given m/z, the width
                        corresponds to he given format, e.g. 'triangle.MS_precision'
                    *   'label.hoverinfo':
                        Label string appears as plotly hover info
                    *   'label.linear'   (top, medium or bottom)
                    *   'label.spline'   (top, medium or bottom)
                    *   'label.sticks'
                    *   'label.triangle' (small, medium or big)
            mz_range (list): Boundaries that should be added to the current
                plot
            int_range (list): Boundaries that should be added to the current
                plot
            opacity (float): opacity of the data points
            dash (str): type of line ('solid', 'dash', 'longdash', 'dot', 'dashdot', 'longdashdot')
            name (str): name of data in legend
            plot_num (int): Add data to plot[plot_num]
            title (str): an optional title that will be printed above the plot

        Note:
            The mz_range and int_range in the add() function sets the limits of datapoints
            added to the plot. This is in contrast to defining a range in the layout, which
            only defines the area that is depicted (i.e. the zoom) but still adds the datapoints to the plot
            (can be seen by zooming out).

        """
        if len(self.plots) == 0:
            self.new_plot(title=title)
        if mz_range is None:
            mz_range = [-float("Inf"), float("Inf")]
        if int_range is None:
            int_range = [-float("Inf"), float("Inf")]
        if len(self.x_max) < len(self.plots):
            self.x_max.append(mz_range[1])
            self.x_min.append(mz_range[0])
        if len(self.y_max) < len(self.plots):
            self.y_max.append(int_range[1])
            self.y_min.append(int_range[0])

        # Init variables
        filling = None
        mode = "lines"
        x_values = []
        y_values = []
        txt = []

        style = style.split(".")
        MS_precision = float(self.MS_precisions[plot_num])

        if len(style) == 3:
            pos = style[2]
            available_pos = [
                "MS_precision",
                "micro",
                "tiny",
                "small",
                "medium",
                "big",
                "top",
                "bottom",
            ]
            if pos not in available_pos:
                raise Exception(
                    "Position must one of the following: {0}".format(available_pos)
                )
        else:
            pos = None
        if style[0] == "label":
            mode = "text+lines"
            if len(data[0]) < 3:
                raise Exception(
                    """
                    Must have at least (mz, i, annotation) in data
                    when using labels
                    """
                )
            if style[1] == "hoverinfo":
                shape = "linear"
                mode = None
                filling = None
                for data_tuple in data:
                    x_values.append(data_tuple[0])
                    y_values.append(data_tuple[1])
                    txt.append(data_tuple[2])

            elif style[1] == "sticks":
                shape = "linear"
                filling = "tozeroy"
                for x in data:
                    y_pos = x[1]
                    x_values += x[0], x[0], x[0], None
                    y_values += 0.0, y_pos, 0.0, None
                    txt += None, x[2], None, None

            elif style[1] == "triangle":
                if not pos:
                    pos = "medium"
                triangle_pos_list = [
                    "MS_precision",
                    "micro",
                    "tiny",
                    "small",
                    "medium",
                    "big",
                ]
                if pos not in triangle_pos_list:
                    raise Exception("Position must be in {0}".format(triangle_pos_list))

                shape = "linear"
                filling = "tozeroy"
                for x in data:
                    x_max = self.x_max[plot_num]
                    y_max = x[1]
                    y_values += 0.0, y_max, 0.0, None
                    txt += None, x[2], None, None

                    if pos == "MS_precision":
                        x_values += (
                            x[0] - (x[0] * MS_precision),
                            x[0],
                            x[0] + (x[0] * MS_precision),
                            None,
                        )
                        continue
                    elif pos == "micro":
                        rel_width = 1 / float(10000)
                    elif pos == "tiny":
                        rel_width = 1 / float(2000)
                    elif pos == "small":
                        rel_width = 1 / float(200)
                    elif pos == "medium":
                        rel_width = 1 / float(100)
                    elif pos == "big":
                        rel_width = 1 / float(50)
                    x_values += (
                        x[0] - (x_max * rel_width),
                        x[0],
                        x[0] + (x_max * rel_width),
                        None,
                    )

            elif style[1] == "spline":
                mode = "lines+markers+text"
                shape = "spline"
                txt = []
                if not pos:
                    pos = "top"
                for x in data:
                    if x[2] == "max_intensity":
                        y_max = self.y_max[plot_num]
                    else:
                        y_max = x[1]
                    if pos == "top":
                        y_pos = y_max
                        offset = y_max + (y_max * 0.1)

                    elif pos == "medium":
                        print(
                            """'
                            {0}
                            is not working atm for
                            {1}
                            """.format(
                                pos, style
                            )
                        )
                        sys.exit(0)
                        y_pos = x[2] / 2
                        offset = "__splineOffset__"

                    elif pos == "bottom":
                        y_pos = 0.0
                        offset = "-__splineOffset__0"

                    x_values += x[0], (x[0] + x[1]) / 2, x[1], None
                    y_values += y_pos, str(offset), y_pos, None
                    txt += None, x[3], None, None

            elif style[1] == "linear":
                shape = "linear"
                if not pos:
                    pos = "bottom"
                for x in data:
                    if x[2] == "max_intensity":
                        y_max = self.y_max[plot_num]
                    else:
                        y_max = x[1]
                    if pos == "top":
                        y_pos = y_max
                        offset = y_max + (y_max * 0.1)
                    elif pos == "medium":
                        print(
                            """'
                            {0}
                            is not working atm for
                            {1}
                            """.format(
                                pos, style
                            )
                        )
                        sys.exit(0)
                        y_pos = x[2] / 2
                        offset = "+__splineOffset__"
                    elif pos == "bottom":
                        y_pos = 0.0
                        offset = "-__splineOffset__0"

                    x_values += x[0], (x[0] + x[1]) / 2, x[1], None
                    y_values += str(offset), str(offset), str(offset), None
                    txt += None, x[3], None, None

            # elif style[1] == 'lines':
            #         shape = 'linear'
            #         filling = 'tozeroy'
            #         for x in data:
            #             y_pos = 'self.y_max[i]'
            #             x_values += x[0], x[0], x[0], None
            #             y_values += .0, y_pos, .0, None
            #             txt     += None, x[2], None, None

            else:
                raise Exception(
                    """
                    Unknown label type
                    Currently supported are:
                    -> linear
                    -> spline
                    -> sticks
                    -> triangle
                    """
                )

        elif style[0] in ["sticks", "triangle", "lines", "points"]:
            x_vals = [
                mz
                for mz, i in data
                if mz_range[0] <= mz <= mz_range[1]
                and int_range[0] <= i <= int_range[1]
            ]
            y_vals = [
                i
                for mz, i in data
                if mz_range[0] <= mz <= mz_range[1]
                and int_range[0] <= i <= int_range[1]
            ]
            y_max = max(y_vals)
            y_min = min(y_vals)
            x_max = max(x_vals)
            x_min = min(x_vals)

            if self.x_max[plot_num] == float("Inf") or self.x_max[plot_num] < x_max:
                self.x_max[plot_num] = x_max
            if self.x_min[plot_num] == -float("Inf") or self.x_min[plot_num] > x_min:
                self.x_min[plot_num] = x_min
            if self.y_max[plot_num] == float("Inf") or self.y_max[plot_num] < y_max:
                self.y_max[plot_num] = y_max
            if self.y_min[plot_num] == -float("Inf") or self.y_min[plot_num] > y_min:
                self.y_min[plot_num] = y_min

            if style[0] == "sticks":
                shape = "linear"
                mode = "lines"
                filling = "tozeroy"
                for x in zip(x_vals, y_vals):
                    y_pos = x[1]
                    x_values += x[0], x[0], x[0], None
                    y_values += 0.0, y_pos, 0.0, None

            elif style[0] == "triangle":
                if len(style) == 2:
                    pos = style[1]
                else:
                    pos = "medium"
                triangle_pos_list = [
                    "MS_precision",
                    "micro",
                    "tiny",
                    "small",
                    "medium",
                    "big",
                ]
                if pos not in triangle_pos_list:
                    raise Exception("Position must be in {0}".format(triangle_pos_list))
                shape = "linear"
                filling = "tozeroy"
                for x in zip(x_vals, y_vals):
                    x_max = self.x_max[plot_num]
                    y_pos = x[1]
                    y_values += 0.0, y_pos, 0.0, None
                    if pos == "MS_precision":
                        x_values += (
                            x[0] - (x[0] * MS_precision),
                            x[0],
                            x[0] + (x[0] * MS_precision),
                            None,
                        )
                        continue
                    elif pos == "micro":
                        rel_width = 1 / float(10000)
                    elif pos == "tiny":
                        rel_width = 1 / float(2000)
                    elif pos == "small":
                        rel_width = 1 / float(200)
                    elif pos == "medium":
                        rel_width = 1 / float(100)
                    elif pos == "big":
                        rel_width = 1 / float(50)
                    x_values += (
                        x[0] - (x_max * rel_width),
                        x[0],
                        x[0] + (x_max * rel_width),
                        None,
                    )

            elif style[0] == "lines":
                mode = "lines"
                shape = "linear"
                for x in zip(x_vals, y_vals):
                    x_values.append(x[0])
                    y_values.append(x[1])

            elif style[0] == "points":
                mode = "markers"
                shape = "linear"
                x_values = x_vals
                y_values = y_vals

        else:
            raise Exception(
                """
                Invalid plotting style
                Currently supported are:
                -> lines
                -> points
                -> sticks
                -> triangle
                """
            )

        trace = go.Scatter(
            {
                "x": x_values,
                "y": y_values,
                "text": txt,
                "textfont": {"family": "Helvetica", "size": 10, "color": "#000000"},
                "textposition": "top center",
                "visible": True,
                "marker": {
                    "size": 10,
                    "color": "rgba({0},{1},{2},{3})".format(
                        color[0], color[1], color[2], opacity
                    ),
                },
                "mode": mode,
                "name": name,
                "line": {
                    "color": "rgba({0},{1},{2},{3})".format(
                        color[0], color[1], color[2], opacity
                    ),
                    "width": self.style_options["line.width"],
                    "shape": shape,
                    "dash": dash,
                },
                "fill": filling,
                "fillcolor": "rgba({0},{1},{2},{3})".format(
                    color[0], color[1], color[2], opacity
                ),
                "opacity": opacity,
            }
        )

        self.plots[plot_num].append(trace)
        return trace

    def info(self):
        """
        Prints summary about the plotting factory, i.e. how many plots and how
        many datasets per plot.
        """
        print(
            """
            Factory holds {0} unique plots
            """.format(
                len(self.plots)
            )
        )
        for i, plot in enumerate(self.plots):
            print("\t\tPlot {0} holds {1} unique datasets".format(i, len(plot)))
            for j, dataset in enumerate(plot):
                print(
                    "\t\t\tDataset {0} holds {1} datapoints".format(
                        j, len(dataset["x"])
                    )
                )

        print()
        return

    def save(
        self, filename=None, mz_range=None, int_range=None, layout=None, write_pdf=False
    ):
        """
        Saves all plots and their data points that have been added to the
        plotFactory.

        Keyword Arguments:
            filename (str): Name for the output file. Default = "spectrum_plot.html"
            mz_range (list): m/z range which should be considered [start, end].
                Default = None
            int_range (list): intensity range which should be considered [min, max].
                Default = None
            layout (dict): dictionary containing layout information in plotly style
            write_pdf (bool): Set "True" in order to save plots as pdf file (on Unix systems,
                this requires Orca to be installed)

        Note:
            mz_range and int_range defined here will be applied to all subplots in
            the current plot, i.e. ranges defined when adding a subplot will be
            overwritten. To avoid this, a list of lists can be given in the order
            corresponding to the subplots.
        """
        plot_number = len(self.plots)
        rows, cols = int(math.ceil(plot_number / float(2))), 2
        if plot_number % 2 == 0:
            my_figure = subplots.make_subplots(
                rows=rows,
                cols=cols,
                vertical_spacing=0.6 / rows,
                subplot_titles=self.titles,
            )
        else:
            specs = [[{}, {}] for x in range(rows - 1)]
            specs.append([{"colspan": 2}, None])
            my_figure = subplots.make_subplots(
                rows=rows,
                cols=cols,
                vertical_spacing=0.6 / rows,
                specs=specs,
                subplot_titles=self.titles,
            )

        for i, plot in enumerate(self.plots):
            print(int(math.floor((i / 2) + 1)), (i % 2) + 1)
            for j, trace in enumerate(plot):
                trace["y"] = [
                    self.function_mapper[x](i) if x in self.function_mapper else x
                    for x in trace["y"]
                ]
                my_figure.append_trace(trace, int(math.floor((i / 2) + 1)), (i % 2) + 1)

        for i in range(plot_number):
            xaxis_key = "xaxis{0}".format(i + 1)
            if mz_range:
                if mz_range[0] == list or mz_range[0] == tuple:
                    # if list and not list of lists, apply same for all
                    my_figure["layout"][xaxis_key].update(range=mz_range[i])
                    my_figure["layout"][xaxis_key]["autorange"] = False
                elif mz_range[0] == int or mz_range[0] == float:
                    my_figure["layout"][xaxis_key]["range"] = mz_range
                    my_figure["layout"][xaxis_key]["autorange"] = False
            yaxis_key = "yaxis{0}".format(i + 1)
            if int_range:
                if int_range[0] == list or int_range[0] == tuple:
                    # if list and not list of lists, apply same for all
                    my_figure["layout"][yaxis_key].update(range=int_range[i])
                    my_figure["layout"][yaxis_key]["autorange"] = False
                elif int_range[0] == int or int_range[0] == float:
                    my_figure["layout"][yaxis_key]["range"] = int_range
                    my_figure["layout"][yaxis_key]["autorange"] = False

            my_figure["layout"][xaxis_key].update(title="m/z ")
            my_figure["layout"][yaxis_key].update(title="Intensity")
            my_figure["layout"][xaxis_key].update(
                titlefont={"color": "#000000", "family": "Helvetica", "size": 18}
            )
            my_figure["layout"][yaxis_key].update(
                titlefont={"color": "#000000", "family": "Helvetica", "size": 18}
            )

        my_figure["layout"]["legend"].update(font={"size": 10, "color": "#FF0000"})
        if layout:
            my_figure["layout"].update(layout)
        if self.filename is None:
            _filename = "spectrum_plot.html"
        else:
            _filename = self.filename

        if filename is not None:
            # save fkt name definition overrules original filename
            _filename = filename

        # plt.plot(my_figure, filename=_filename, auto_open=False)
        plt.io.write_html(my_figure, _filename)
        if write_pdf:
            plt.io.write_image(my_figure, _filename.replace(".html", ".pdf"))
        return

    def get_data(self):
        """
        Return data and layout in JSON format.

        Returns:
            plots (dict): JSON compatible python dict

        """
        for i, plot in enumerate(self.plots):
            for j, trace in enumerate(plot):
                self.plots[i][j]["y"] = [
                    self.function_mapper[x](i) if x in self.function_mapper else x
                    for x in trace["y"]
                ]
        return self.plots


if __name__ == "__main__":
    print(__doc__)



# -*- coding: utf-8 -*-
# encoding: utf-8
"""
Plotting functions for pymzML.
The Factory object can hold several plots with several data traces each.
The plot will be rendered as interactive plotly plots.
"""

# Python mzML module - pymzml
# Copyright (C) 2010-2019 M. Kösters, C. Fufezan
#     The MIT License (MIT)

#     Permission is hereby granted, free of charge, to any person obtaining a copy
#     of this software and associated documentation files (the "Software"), to deal
#     in the Software without restriction, including without limitation the rights
#     to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#     copies of the Software, and to permit persons to whom the Software is
#     furnished to do so, subject to the following conditions:

#     The above copyright notice and this permission notice shall be included in all
#     copies or substantial portions of the Software.

#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#     IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#     FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#     AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#     LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#     OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#     SOFTWARE.

import math
import plotly.offline as plt
import plotly.graph_objs as go
from plotly import tools


class Factory(object):
    def __init__(self, filename="spectrum_plot.html"):
        """
        Class to plot pymzml.spec.Spectrum using plotly.

        Args:
            filename (str): Name for the output file. Default = "spectrum_plot.html"

        Example:

            >>> import pymzml, get_example_file
            >>> mzMLFile = 'profile-mass-spectrum.mzml'
            >>> example_file = get_example_file.open_example(mzMLFile)
            >>> run = pymzml.run.Run("../mzML_example_files/"+mzMLFile,
            ...     precisionMSn = 250e-6
            ... )
            >>> p = pymzml.plot.Factory()
            >>> for spec in run:
            >>>     p.new_plot()
            >>>     p.add(spec.peaks, color=(200,00,00), style='sticks')
            >>>     p.add(spec.centroided_peaks, color=(00,00,00), style='sticks')
            >>>     p.add(spec.reprofiled_peaks, color=(00,255,00), style='sticks')
            >>>     p.save(
            ...         filename="output/plotAspect.xhtml",
            ...         mz_range = (745.2,745.6)
            ...     )

        NOTE:
            This should go into example_script/plot_spectrum.py

        """
        self.filename = filename
        self.plots = []
        self.titles = []
        self.lookup = dict()
        self.y_max = []
        self.x_max = []
        self.offset = 1
        self.precisions = []
        self.styles = ["sticks", "triangle", "spline", "linear"]
        self.widths = ["small", "medium", "big"]
        self.style_parameters = {
            "sticks": {
                "mode": "lines+text",
                "shape": "linear",
                "filling": "tozeroy",
                "width": lambda n: self.precisions[n],
            },
            "triangle": {
                "mode": "lines+text",
                "shape": "linear",
                "filling": "tozeroy",
                "width": lambda n: self.scalings[n],
            },
            "lines": {"mode": "lines+text", "shape": "spline", "filling": None},
            "points": {"mode": "markers", "shape": "linear", "filling": None},
        }

    def new_plot(self, precision="5e-6", title=None):
        """
        Add new plot to the plotting Factory.

        Args:
            header (str): an optional title that will be printed above the plot
            mz_range (tuple): Boundaries of the new plot
            normalize (boolean): whether or not the individal data sets
                are normalized in the plot
            precision (float): measuring precision used in handler.
                Default 5e-6.
        """
        self.precisions.append(precision)
        self.plots.append([])
        return

    def add(
        self,
        data,
        color=(0, 0, 0),
        style="sticks",
        mz_range=None,
        opacity=0.8,
        name=None,
        plot_num=-1,
        title=None,
    ):
        """
        Add data to the graph.

        Arguments:
            data (list): The data added to the graph. Must be list of
                tuples, like (mz,i) or (mz1, mz2, i, string)
                style (str): plotting style. Default = "sticks".\n
                Currently supported styles are:\n
                    *   'sticks'\n
                    *   'triangle' (small, medium or big)\n
                    *   'spline'   (top, medium or bottom)\n
                    *   'linear'   (top, medium or bottom)\n
            color (tuple): color encoded in RGB. Default = (0,0,0)
            mz_range (tuple): Boundaries that should be added to the current
                plot
            opacity (float): opacity of the data points\n
            name (str): name of data in legend\n
            plot_num (int): Add data to plot[plot_num]\n

        """
        style_attribs = style.split(".")
        assert (
            len(style_attribs) == 3
        ), "Style must set datatype, plotting style and width:\n{0}".format(
            style_attribs
        )

        if len(self.plots) == 0:
            self.new_plot(title=title)

        precision = float(self.precisions[plot_num])

        if style_attribs[0] == "label":
            as_anno = True
        elif style_attribs[0] == "data":
            as_anno = False
        else:
            raise Exception("Style must be declare trace as data or annotation")

        # use numpy arrays
        x_values = []
        y_values = []
        txt_values = []

        if as_anno:
            for vars in data:
                mz, i = vars[0], vars[1]
                txt = vars[2]
                x_values.extend(
                    [
                        mz
                        - (
                            mz * precision
                        ),  # self.style_parameters[style_attribs[1]]['width'])(plot_num)
                        mz,
                        mz
                        + (
                            mz * precision
                        ),  # self.style_parameters[style_attribs[1]]['width'])(plot_num)
                        None,
                    ]
                )
                y_values.extend(
                    [
                        0,  # y pos for spline, offset for linear, also via dict grab
                        i,  # offset for linear
                        0,  # ypos for spline, offset for linear, also via dict grab
                        None,
                    ]
                )
                txt_values.extend([None, txt, None, None])
        else:
            for mz, i in data:
                x_values.extend(
                    [mz - (mz * precision), mz, mz + (mz * precision), None]
                )
                y_values.extend([0.0, i, 0.0, None])

        data = go.Scatter(
            {
                "x": x_values,
                "y": y_values,
                "text": txt_values,
                "textfont": {"family": "Helvetica", "size": 10, "color": "#000000"},
                "textposition": "top center",
                "visible": "True",
                "marker": {"size": 10},
                "mode": self.style_parameters[style_attribs[1]]["mode"],
                "name": name,
                "line": {
                    "color": "rgb" + str(color),
                    "width": 1,
                    "shape": self.style_parameters[style_attribs[1]]["shape"],
                },
                "fill": self.style_parameters[style_attribs[1]]["filling"],
                "fillcolor": {
                    "color": "rgba" + str((color[0], color[1], color[2], opacity))
                },
                "opacity": opacity,
            }
        )

        self.plots[plot_num].append(data)
        return

    def info(self):
        """
        Prints summary about the plotting factory, i.e.how many plots and how
        many datasets per plot.
        """
        print(
            """
        Factory holds {0} unique plots""".format(
                len(self.plots)
            )
        )
        for i, plot in enumerate(self.plots):
            print("\t\tPlot {0} holds {1} unique datasets".format(i, len(plot)))
            for j, dataset in enumerate(plot):
                print(
                    "\t\t\tDataset {0} holds {1} datapoints".format(
                        j, len(dataset["x"])
                    )
                )

        print()
        return

    def get_data(self):
        """
        Return data and layout in JSON format.

        Returns:
            plots (dict): JSON compatible python dict
        """
        for i, plot in enumerate(self.plots):
            for j, trace in enumerate(plot):
                self.plots[i][j]["y"] = [
                    self.function_mapper[x](i) if x in self.function_mapper else x
                    for x in trace["y"]
                ]
        return self.plots

    def save(self, filename=None, xLimits=None):
        """
        Saves all plots and their data points that have been added to the
        plotFactory.

        Args:
        filename (str): Name for the output file. Default = "spectrum_plot.html"
        mz_range (tuple): m/z range which should be considered [start, end].
            Default = None
        """
        plot_number = len(self.plots)

        rows, cols = int(math.ceil(plot_number / float(2))), 2

        if plot_number % 2 == 0:
            my_figure = tools.make_subplots(
                rows=rows, cols=cols, vertical_spacing=0.6 / rows
            )
        else:
            specs = [[{}, {}] for x in range(rows - 1)]
            specs.append([{"colspan": 2}, None])
            my_figure = tools.make_subplots(
                rows=rows,
                cols=cols,
                vertical_spacing=0.6 / rows,
                specs=specs,
                subplot_titles=self.titles,
            )

        for i, plot in enumerate(self.plots):
            for j, trace in enumerate(plot):
                my_figure.append_trace(trace, int(math.floor((i / 2) + 1)), (i % 2) + 1)

        for i in range(plot_number):
            if xLimits:
                my_figure["layout"]["xaxis" + str(i + 1)].update(range=xLimits[i])
            my_figure["layout"]["xaxis" + str(i + 1)].update(title="m/z ")
            my_figure["layout"]["yaxis" + str(i + 1)].update(title="Intensity")
            my_figure["layout"]["xaxis" + str(i + 1)].update(
                titlefont={"color": "#000000", "family": "Helvetica", "size": "18"}
            )
            my_figure["layout"]["yaxis" + str(i + 1)].update(
                titlefont={"color": "#000000", "family": "Helvetica", "size": "18"}
            )

        my_figure["layout"]["legend"].update(font={"size": 10, "color": "#FF0000"})

        if self.filename is None:
            _filename = "spectrum_plot.html"
        else:
            _filename = self.filename

        if filename is not None:
            # save fkt name definition overrules original filename
            _filename = filename

        plt.plot(my_figure, filename=_filename, auto_open=False)
        return


if __name__ == "__main__":
    data = [(1, 10), (2, 20)]
    anno = [(1, 10, "blub"), (2, 20, "bla")]
    Fac = Factory()

    print("'style='data.sticks.medium'")
    Fac.add(data, style="data.sticks.medium")
    print()
    print("style='data.triangle.small'")
    Fac.add(data, style="data.triangle.small")
    print()
    Fac.add(anno, style="label.sticks.medium")
    Fac.save(filename="tmp.html")



#!/usr/bin/env python3
"""
Classes to encode and decode :py:attr:`~pymzml.spec.Spectrum.mz` and
:py:attr:`~pymzml.spec.Spectrum.i` values.

@author M. Kösters, C. Fufezan
"""
import warnings
import zlib
from base64 import b64decode as b64dec
from multiprocessing import Pool

import numpy as np

# Global PyNump decoder
try:
    # try to import c-accelerated Numpress decoding
    import pynumpress

    MSDecoder = pynumpress
except ImportError:
    # fall back to python-only implementation of numpress decoding
    import pymzml.ms_numpress

    warnings.warn(
        "Cython PyNumpress is not installed; falling back to slower, python-only version",
        ImportWarning,
    )
    MSDecoder = pymzml.ms_numpress.MSNumpress()


def _decode(data, comp, d_array_length, f_type, d_type):
    """
    Decode ms-numpress, b64 and/or zlib compressed data.

    Args:
        data (str): compressed data
        comp (str): compression method
        d_array_length (int): length of the uncompressed data array
        fType (str): float type (32 or 64 bit)
        d_type (str): type of data (mz, i, or time)

    Returns:
        result (tuple(str, list)): tuple containing the datatype and the
        decompressed data list.
    """
    if f_type == "32-bit float":
        f_type = np.float32
    elif f_type == "64-bit float":
        f_type = np.float64
    else:
        f_type = None

    decoded_data = b64dec(data)
    if "zlib" in comp or "zlib compression" in comp:
        decoded_data = zlib.decompress(decoded_data)

    if (
        "ms-np-linear" in comp
        or "ms-np-pic" in comp
        or "ms-np-slof" in comp
        or "MS-Numpress linear prediction compression" in comp
        or "MS-Numpress short logged float compression" in comp
    ):
        result = []
        # start ms numpress decoder globally?
        if (
            "ms-np-linear" in comp
            or "MS-Numpress linear prediction compression" in comp
        ):
            result = MSDecoder.decodeLinear(decoded_data)
        elif "ms-np-pic" in comp:
            result = MSDecoder.decode_pic(decoded_data)
        elif (
            "ms-np-slof" in comp or "MS-Numpress short logged float compression" in comp
        ):
            result = MSDecoder.decode_slof(decoded_data)
        return (d_type, result)

    array = np.fromstring(decoded_data, f_type)
    return (d_type, array)


class Decoder:
    """
    Decoder class to enable parallel decoding of peaks.

    Keyword Args:
        nb_worker(int): number of pool workers to use. Defaults to 2.
    """

    def __init__(self, nb_workers=2):
        """ """
        self._mz = None
        self._i = None

    # @profile
    def pool_decode(self, data, callback):
        """
        Decode mz and i values in parallel.

        Args:
            data (): ...

        Keyword Args:
            callback (:obj:`func`): Callback function to call if decoding is
                finished. Should be :py:meth:`~pymzml.spec.Spectrum._register`.
        """
        ZE_POOL = Pool(processes=2)

        ZE_POOL.starmap(_decode, data)

    def _error_callback(self, result):
        """ """
        raise Exception("Failed with error:\n{0}".format(result))


if __name__ == "__main__":
    print(__doc__)



import unittest
import struct
import math

try:
    import numpy as np
except:
    np = None

import pynumpress


@unittest.skipIf(np is None, "Numpy is required for this test.")
class test_MSNumpress(unittest.TestCase):

    """
    unittest for MSNumpress en- and decoding
    """

    def setUp(self):
        """ """
        self.mz_data = np.asarray([100.1, 100.01, 100.001, 100.0001], dtype=np.float64)
        self.i_data = np.array([5e6, 6.5e5, 2e6, 12e6])
        self.i_slof_data = np.array([1.0, 2.0, 4.0])
        self.integer = 23
        # self._encoded_int = bytes([ 0x06, 0x07, 0x01 ])
        # pynumpress = MSNumpress([])
        self.fixed_point = 10000

    def test_encode_slof(self):
        fp = pynumpress.optimal_slof_fixed_point(self.i_slof_data)
        encoded_array = pynumpress.encode_slof(
            np.asarray(self.i_slof_data, dtype=np.float64), fp
        )

        self.assertEqual(len(encoded_array), 14)

        self.assertEqual(encoded_array[0], 0x40)
        self.assertEqual(encoded_array[1], 0xE3)
        self.assertEqual(encoded_array[2], 0xE1)
        self.assertEqual(encoded_array[3], 0xE0)
        self.assertEqual(encoded_array[4], 0x0)
        self.assertEqual(encoded_array[5], 0x0)
        self.assertEqual(encoded_array[6], 0x0)
        self.assertEqual(encoded_array[7], 0x0)
        self.assertEqual(encoded_array[8], 0x40)
        self.assertEqual(encoded_array[9], 0x6E)
        self.assertEqual(encoded_array[10], 0xBE)
        self.assertEqual(encoded_array[11], 0xAE)
        self.assertEqual(encoded_array[12], 0xFF)
        self.assertEqual(encoded_array[13], 0xFF)

    def test_decode_slof(self):
        test_array = [
            0x40,
            0xC3,
            0x88,
            0x0,
            0x0,
            0x0,
            0x0,
            0x0,
            0x13,
            0x1B,
            0xEA,
            0x2A,
            0xDE,
            0x3E,
        ]
        test_array = bytearray(test_array)
        # pynumpress.encoded_data = test_array
        decoded_array = pynumpress.decode_slof(test_array)
        for i, dec in enumerate(decoded_array):
            self.assertAlmostEqual(dec, self.i_slof_data[i], places=2)
        # self.assertCountEqual(decoded_array, self.i_slof_data)

    def test_encode_decode_slof(self):
        """ """
        # pynumpress.decoded_data = self.i_slof_data
        print(self.i_slof_data)
        fp = pynumpress.optimal_slof_fixed_point(self.i_slof_data)
        encoded_array = pynumpress.encode_slof(
            np.asarray(self.i_slof_data, dtype=np.float64), fp
        )
        decoded_array = pynumpress.decode_slof(encoded_array)
        for i, dec in enumerate(decoded_array):
            self.assertAlmostEqual(dec, self.i_slof_data[i], places=2)

    def test_encode_pic_i_data(self):
        encoded_array = pynumpress.encode_pic(np.asarray(self.i_data, dtype=np.float64))

        self.assertEqual(
            len(encoded_array), 14, msg="{}".format([hex(x) for x in encoded_array])
        )

        self.assertEqual(encoded_array[0], 0x20)
        self.assertEqual(encoded_array[1], 0x4B)
        self.assertEqual(encoded_array[2], 0x4C)
        self.assertEqual(encoded_array[3], 0x43)
        self.assertEqual(encoded_array[4], 0x1)
        self.assertEqual(encoded_array[5], 0xBE)
        self.assertEqual(encoded_array[6], 0x92)
        self.assertEqual(encoded_array[7], 0x8)
        self.assertEqual(encoded_array[8], 0x48)
        self.assertEqual(encoded_array[9], 0xE1)
        self.assertEqual(encoded_array[10], 0x20)
        self.assertEqual(encoded_array[11], 0xB)
        self.assertEqual(encoded_array[12], 0x17)
        self.assertEqual(encoded_array[13], 0xB0)

    def test_decode_pic_i_data(self):
        test_array = [
            0x20,
            0x4B,
            0x4C,
            0x43,
            0x1,
            0xBE,
            0x92,
            0x8,
            0x48,
            0xE1,
            0x20,
            0xB,
            0x17,
            0xB0,
        ]
        decoded_array = pynumpress.decode_pic(test_array)
        self.assertCountEqual(
            decoded_array,
            self.i_data,
            msg="{}\n{}".format([x for x in decoded_array], [x for x in self.i_data]),
        )

    def test_encode_decode_pic(self):
        """ """
        encoded_array = pynumpress.encode_pic(
            np.asarray(self.i_slof_data, dtype=np.float64)
        )
        decoded_array = pynumpress.decode_pic(encoded_array)
        self.assertCountEqual(self.i_slof_data, decoded_array)

    def test_encode_linear(self):
        decoded_data = self.mz_data
        fp = pynumpress.optimal_linear_fixed_point(self.mz_data)
        encoded_array = pynumpress.encode_linear(decoded_data, fp)
        self.assertEqual(
            len(encoded_array),
            23,
            msg="{0}\n{1}".format(
                [hex(x) for x in encoded_array],
                [
                    "0x41",
                    "0x74",
                    "0x75",
                    "0xa4",
                    "0x70",
                    "0x0",
                    "0x0",
                    "0x0",
                    "0xf6",
                    "0xff",
                    "0xff",
                    "0x7f",
                    "0xc2",
                    "0x89",
                    "0xe2",
                    "0x7f",
                    "0x2b",
                    "0xf3",
                    "0x8a",
                    "0x13",
                    "0xdc",
                    "0x6a",
                    "0x20",
                ],
            ),
        )

        # assert first value
        self.assertEqual(
            0xFF & encoded_array[8],
            0xF6,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                8, hex(0xFF & encoded_array[8]), hex(0xF6)
            ),
        )

        self.assertEqual(
            0xFF & encoded_array[9],
            0xFF,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                9, hex(0xFF & encoded_array[9]), hex(0xFF)
            ),
        )
        self.assertEqual(
            0xFF & encoded_array[10],
            0xFF,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                10, hex(0xFF & encoded_array[10]), hex(0xFF)
            ),
        )
        self.assertEqual(
            0xFF & encoded_array[11],
            0x7F,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                11, hex(0xFF & encoded_array[11]), hex(0x7F)
            ),
        )

        # assert second value
        self.assertEqual(
            0xFF & encoded_array[12],
            0xC2,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                12, hex(0xFF & encoded_array[12]), hex(0xC2)
            ),
        )
        self.assertEqual(
            0xFF & encoded_array[13],
            0x89,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                13, hex(0xFF & encoded_array[13]), hex(0x89)
            ),
        )
        self.assertEqual(
            0xFF & encoded_array[14],
            0xE2,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                14, hex(0xFF & encoded_array[14]), hex(0xE2)
            ),
        )
        self.assertEqual(
            0xFF & encoded_array[15],
            0x7F,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                15, hex(0xFF & encoded_array[15]), hex(0x7F)
            ),
        )

        # assert third value
        self.assertEqual(
            0xFF & encoded_array[16],
            0x2B,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                16, hex(0xFF & encoded_array[16]), hex(0x2B)
            ),
        )
        self.assertEqual(
            0xFF & encoded_array[17],
            0xF3,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                17, hex(0xFF & encoded_array[17]), hex(0xC2)
            ),
        )
        self.assertEqual(
            0xFF & encoded_array[18],
            0x8A,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                18, hex(0xFF & encoded_array[18]), hex(0x8A)
            ),
        )
        self.assertEqual(
            0xFF & encoded_array[19],
            0x13,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                19, hex(0xFF & encoded_array[19]), hex(0x13)
            ),
        )

        # assert fourth value
        self.assertEqual(
            0xFF & encoded_array[20],
            0xDC,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                19, hex(0xFF & encoded_array[20]), hex(0xDC)
            ),
        )
        self.assertEqual(
            0xFF & encoded_array[21],
            0x6A,
            msg="Fail in value at pos {0}: {1} != {2}".format(
                19, hex(0xFF & encoded_array[21]), hex(0x6A)
            ),
        )
        # self.assertEqual(
        #     0xff & encoded_array[22],
        #     0x20,
        #     msg='Fail in value at pos {0}: {1} != {2}'.format(
        #         19,
        #         hex(0xff & encoded_array[22]),
        #         hex(0x20)
        #     )
        # )

    def test_decode_linear(self):

        # ouput of  >>> PyMSNumpress.encode_linear(self.mz_data, enc,
        # self.fixed_point)
        encoded_array = [
            65,
            116,
            117,
            164,
            112,
            0,
            0,
            0,
            246,
            255,
            255,
            127,
            194,
            137,
            226,
            127,
            43,
            243,
            138,
            19,
            220,
            106,
            32,
        ]

        # make bytearray from hex vals
        encoded_array = np.array(encoded_array, dtype=np.dtype("B"))

        # decode bytearray to numpy array
        decoded_array = pynumpress.decode_linear(encoded_array)
        self.assertIsInstance(decoded_array, np.ndarray)
        self.assertEqual(len(decoded_array), len(self.mz_data))
        for i in range(len(decoded_array)):
            self.assertAlmostEqual(decoded_array[i], self.mz_data[i], places=i + 1)

    def test_encode_decode_linear(self):
        """ """
        test_array = [
            100.00066,
            100.00217,
            100.00368,
            100.00519,
            111.73335,
            111.73513,
            111.73692,
            111.7387,
            111.74049,
            111.74227,
            111.74406,
            111.74584,
            111.74763,
            111.74941,
            111.7512,
            111.75298,
            111.75477,
            112.00694,
            112.00873,
            112.01052,
            112.01231,
            112.0141,
            112.01589,
            112.01768,
        ]
        test_array = np.asarray(test_array, dtype=np.float64)
        fp = pynumpress.optimal_linear_fixed_point(test_array)
        encoded_array = pynumpress.encode_linear(test_array, fp)
        decoded_array = pynumpress.decode_linear(encoded_array)
        for i in range(len(decoded_array)):
            self.assertAlmostEqual(
                decoded_array[i],
                test_array[i],
                places=4,
                msg="error at pos {0}".format(i),
            )

    def test_encode_decode_self_mz(self):
        mz_data = np.asarray(self.mz_data, dtype=np.float64)
        fp = pynumpress.optimal_linear_fixed_point(mz_data)
        encoded_mz_data = pynumpress.encode_linear(mz_data, fp)
        decoded_mz_data = pynumpress.decode_linear(encoded_mz_data)
        for i in range(len(decoded_mz_data)):
            self.assertAlmostEqual(
                decoded_mz_data[i],
                mz_data[i],
                places=4,
                msg="error at pos {0}".format(i),
            )


if __name__ == "__main__":
    unittest.main(verbosity=3)



#!/usr/bin/env python
import sys
import pymzml


def main(mzml_file):
    """
    Basic example script to access basic run info of an mzML file. Requires a
    mzML file as first command line argument.

    usage:

        ./access_run_info.py <path_to_mzml_file>

    >>> run.info =
            {
                'encoding': 'utf-8',
                 'file_name': '/Users/joe/Dev/pymzml_2.0/tests/data/BSA1.mzML.gz',
                 'file_object': <pymzml.file_interface.FileInterface object at 0x1039a3f28>,
                 'obo_version': '1.1.0',
                 'offset_dict': None,
                 'run_id': 'ru_0',
                 'spectrum_count': 1684,
                 'start_time': '2009-08-09T22:32:31'
             }

    """
    run = pymzml.run.Reader(mzml_file)
    print(
        """
Summary for mzML file:
    {file_name}
Run was measured on {start_time} using obo version {obo_version}
File contains {spectrum_count} spectra
        """.format(
            **run.info
        )
    )


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(main.__doc__)
        exit()
    mzml_file = sys.argv[1]
    main(mzml_file)



#!/usr/bin/env python
"""
Additional functions for converting file etc.

@author M. Kösters
"""

# Python mzML module - pymzml
# Copyright (C) 2010-2019 M. Kösters, C. Fufezan
#     The MIT License (MIT)

#     Permission is hereby granted, free of charge, to any person obtaining a copy
#     of this software and associated documentation files (the "Software"), to deal
#     in the Software without restriction, including without limitation the rights
#     to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#     copies of the Software, and to permit persons to whom the Software is
#     furnished to do so, subject to the following conditions:

#     The above copyright notice and this permission notice shall be included in all
#     copies or substantial portions of the Software.

#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#     IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#     FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#     AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#     LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#     OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#     SOFTWARE.

from pymzml.utils.GSGW import GSGW
import pymzml.regex_patterns as regex_patterns
import re
import gzip


def index_gzip(pathIn, pathOut, max_idx=10000, idx_len=8, verbose=False, comp_str=-1):
    """
    Convert an mzml file (can be gzipped) into an indexed, gzipped mzML file.

    Arguments:
        pathIn (str): path to an mzML input File.
        pathOut (str): path were the index gzip will be created.

    Keyword Arguments:
        max_idx (int): number of indexes which can be saved.
        idx_len (int): character len of on key
        verbose (boolean): print progress while parsing input.
        comp_str(int): compression strength of zlib compression,
            needs to  be 1 <= x <= 9
    """
    if pathIn.endswith("gz"):
        fileOpen = gzip.open
    elif pathIn.lower().endswith("mzml"):
        fileOpen = open
    with GSGW(
        output_path=pathOut,
        max_idx=max_idx,
        max_idx_len=idx_len,
        max_offset_len=idx_len,
        comp_str=comp_str,
    ) as Writer:
        with fileOpen(pathIn, "rt") as Reader:
            data = ""
            for line in Reader:
                if line.strip().startswith("</spectrum>"):
                    data += line
                    Writer.add_data(data, nativeID)
                    if verbose:
                        print("NativeID : {0}".format(nativeID), end="\r")
                    data = ""
                elif line.strip().startswith("<spectrum "):
                    data += line
                    lineID = re.search(regex_patterns.SPECTRUM_TAG_PATTERN, line).group(
                        "index"
                    )
                    nativeID = int(
                        regex_patterns.SPECTRUM_ID_PATTERN.search(lineID).group(1)
                    )

                elif line.strip().startswith("<chromatogram "):
                    data += line
                    nativeID = re.search(
                        regex_patterns.CHROMATOGRAM_ID_PATTERN, line
                    ).group(1)
                    print("found chromatogram")
                elif line.strip().startswith("<spectrumL"):
                    data += line
                    Writer.add_data(data, "Head")
                    if verbose:
                        print("NativeID :", "Head")
                    data = ""
                elif line.strip().startswith("<chromatogramL"):
                    data += line
                    Writer.add_data(data, "junk")
                    if verbose:
                        print("NativeID :", "junk")
                    data = ""
                elif line.strip().startswith("</chromatogram>"):
                    data += line
                    Writer.add_data(data, nativeID)
                    if verbose:
                        print("found chromatogram")
                        print("NativeID: {0}".format(nativeID))
                    data = ""
                else:
                    data += line
            if data:
                Writer.add_data(data, "tail")
                if verbose:
                    print("NativeID :", "tail")
        # print(Writer.index.items())
        Writer.write_index()
    return


def index(pathIn, pathOut, max_idx=10000, idx_len=8, verbose=False, comp_str=-1):
    """
    Convert an mzml file (can be gzipped) into an indexed, gzipped mzML file.

    Arguments:
        pathIn (str): path to input File.
        pathOut (str): path were output should be created.

    Keyword Arguments:
        max_idx (int): number of indexes which can be saved.
        idx_len (int): character len of on key
        verbose (boolean): print progress while parsing input.
        comp_str(int): compression strength of zlib compression,
            needs to  be 1 <= x <= 9
    """
    import gzip

    with GSGW(
        output_path=pathOut,
        max_idx_len=idx_len,
        max_offset_len=idx_len,
        comp_str=comp_str,
    ) as Writer:
        with gzip.open(pathIn, "rt") as Reader:
            data = ""
            for line in Reader:
                if line.strip().startswith("</spectrum>"):
                    data += line
                    Writer.add_data(data, nativeID)
                    if verbose:
                        pass
                    data = ""
                elif line.strip().startswith("<spectrum "):
                    data += line
                    lineID = re.search(regex_patterns.SPECTRUM_TAG_PATTERN, line).group(
                        "index"
                    )
                    nativeID = int(
                        regex_patterns.SPECTRUM_ID_PATTERN.search(lineID).group(0)
                    )
                elif line.strip().startswith("<chromatogram "):
                    data += line
                    nativeID = re.search(
                        regex_patterns.CHROMATOGRAM_ID_PATTERN, line
                    ).group(1)
                elif line.strip().startswith("<spectrumL"):
                    data += line
                    Writer.add_data(data, "Head")
                    if verbose:
                        print("NativeID :", "Head")
                    data = ""
                elif line.strip().startswith("<chromatogramL"):
                    data += line
                    Writer.add_data(data, "junk")
                    if verbose:
                        print("NativeID :", "junk")
                    data = ""
                elif line.strip().startswith("</chromatogram>"):
                    data += line
                    Writer.add_data(data, nativeID)
                    if verbose:
                        print("found chromo")
                        print("NativeID :", nativeID, end="\r")
                    data = ""
                else:
                    data += line
            if data:
                Writer.add_data(data, "tail")
                if verbose:
                    print("NativeID :", "tail")
        Writer.write_index()


def make_obo_mapping(obo, reversed=False):
    # NOT sure what this is for ...
    mapping = {}
    with open(obo) as obo_file:
        for line in obo_file:
            if line.startswith("id: "):
                id = line.split()[-1]
            elif line.startswith("name: "):
                mapping[id] = " ".join(line.split()[1:])
    if reversed:
        mapping = {y: x for x, y in mapping.items()}
    return mapping


if __name__ == "__main__":
    print(__doc__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Interface for gzipped mzML files.
"""

# Python mzML module - pymzml
# Copyright (C) 2010-2019 M. Kösters, C. Fufezan
#     The MIT License (MIT)

#     Permission is hereby granted, free of charge, to any person obtaining a copy
#     of this software and associated documentation files (the "Software"), to deal
#     in the Software without restriction, including without limitation the rights
#     to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#     copies of the Software, and to permit persons to whom the Software is
#     furnished to do so, subject to the following conditions:

#     The above copyright notice and this permission notice shall be included in all
#     copies or substantial portions of the Software.

#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#     IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#     FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#     AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#     LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#     OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#     SOFTWARE.

import codecs
import gzip
from xml.etree.ElementTree import iterparse

from .. import regex_patterns
from .. import spec


class StandardGzip(object):
    def __init__(self, path, encoding):
        """
        Initalize Wrapper object for gzipped mzML files.

        Arguments:
            path (str)     : path to the file
            encoding (str) : encoding of the file
        """
        self.path = path
        self.file_handler = codecs.getreader(encoding)(gzip.open(path))
        self.offset_dict = self._build_index()
        return

    def close(self):
        self.file_handler.close()

    def _build_index(self):
        """
        Cant build index for standard gzip files
        """
        # raise Exception('Cant build index for gzip files')
        pass

    def read(self, size=-1):
        """
        Read binary data from file handler.

        Keyword Arguments:
            size (int): Number of bytes to read from file, -1 to read to end of file

        Returns:
            data (str): byte string of len size of input data
        """
        return self.file_handler.read(size)

    def __getitem__(self, identifier):
        """
        Access the item with id 'identifier' in the file by iterating the xml-tree.

        Arguments:
            identifier (str): native id of the item to access

        Returns:
            data (str): text associated with the given identifier
        """
        old_pos = self.file_handler.tell()
        self.file_handler.seek(0, 0)
        mzml_iter = iter(iterparse(self.file_handler, events=["end"]))
        while True:
            event, element = next(mzml_iter)
            if event == "end":
                if element.tag.endswith("}spectrum"):
                    if (
                        int(
                            regex_patterns.SPECTRUM_ID_PATTERN.search(
                                element.get("id")
                            ).group(1)
                        )
                        == identifier
                    ):
                        self.file_handler.seek(old_pos, 0)
                        return spec.Spectrum(element, measured_precision=5e-6)
                elif element.tag.endswith("}chromatogram"):
                    if element.get("id") == identifier:
                        self.file_handler.seek(old_pos, 0)
                        return spec.Chromatogram(element, measured_precision=5e-6)


if __name__ == "__main__":
    print(__doc__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Interface for uncompressed mzML files.

@author: Manuel Koesters
"""

# Python mzML module - pymzml
# Copyright (C) 2010-2019 M. Kösters, C. Fufezan
#     The MIT License (MIT)

#     Permission is hereby granted, free of charge, to any person obtaining a copy
#     of this software and associated documentation files (the "Software"), to deal
#     in the Software without restriction, including without limitation the rights
#     to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#     copies of the Software, and to permit persons to whom the Software is
#     furnished to do so, subject to the following conditions:

#     The above copyright notice and this permission notice shall be included in all
#     copies or substantial portions of the Software.

#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#     IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#     FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#     AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#     LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#     OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#     SOFTWARE.

import bisect
import codecs
import re
import os
from xml.etree.ElementTree import XML, iterparse

from .. import spec
from .. import regex_patterns


class StandardMzml(object):
    """ """

    def __init__(
        self, path, encoding, build_index_from_scratch=False, index_regex=None
    ):
        """
        Initalize Wrapper object for standard mzML files.

        Arguments:
            path (str)     : path to the file
            encoding (str) : encoding of the file
        """
        self.index_regex = index_regex
        self.path = path
        self.file_handler = self.get_file_handler(encoding)
        self.offset_dict = dict()
        self.spec_open = regex_patterns.SPECTRUM_OPEN_PATTERN
        self.spec_close = regex_patterns.SPECTRUM_CLOSE_PATTERN

        self.seek_list = self._read_extremes()
        self._build_index(from_scratch=build_index_from_scratch)

    def get_binary_file_handler(self):
        return open(self.path, "rb")

    def get_file_handler(self, encoding):
        return codecs.open(self.path, mode="r", encoding=encoding)

    def __getitem__(self, identifier):
        """
        Access the item with id 'identifier'.

        Either use linear, binary or interpolated search.

        Arguments:
            identifier (str): native id of the item to access

        Returns:
            data (str): text associated with the given identifier
        """
        #############################################################################
        # DOES NOT HOLD IF NUMBERS DONT START WITH ONE AND/OR DONT INCREASE BY ONE  #
        # TODO FIXME                                                                #
        #############################################################################

        self.file_handler.seek(0)

        spectrum = None
        if str(identifier).upper() == "TIC":
            # print(str(identifier).upper())
            found = False
            mzmliter = iter(iterparse(self.file_handler, events=["end"]))
            while found is False:
                event, element = next(mzmliter, ("STOP", "STOP"))
                if event == "end":
                    if element.tag.endswith("}chromatogram"):
                        if element.get("id") == "TIC":
                            found = True
                            spectrum = spec.Chromatogram(
                                element, measured_precision=5e-6
                            )
                elif event == "STOP":
                    raise StopIteration

        elif identifier in self.offset_dict:

            start = self.offset_dict[identifier]

            seeker = self.get_binary_file_handler()
            seeker.seek(start[0])
            start, end = self._read_to_spec_end(seeker)

            self.file_handler.seek(start, 0)
            data = self.file_handler.read(end)
            if data.startswith("<spectrum"):
                spectrum = spec.Spectrum(XML(data), measured_precision=5e-6)
            elif data.startswith("<chromatogram"):
                spectrum = spec.Chromatogram(XML(data))
        elif type(identifier) == str:
            return self._search_string_identifier(identifier)
        else:
            spectrum = self._binary_search(identifier)

        return spectrum

    def _binary_search(self, target_index):
        """
        Retrieve spectrum for a given spectrum ID using binary jumps

        Args:
            target_index (int): native id of the spectrum to access

        Returns:
            Spectrum (pymzml.spec.Spectrum): pymzML spectrum


        """
        chunk_size = 12800
        offset_scale = 1
        jump_history = {"forwards": 0, "backwards": 0}
        # This will be used if no spec was found at all during a jump
        # self._average_bytes_per_spec *= 10
        with open(self.path, "rb") as seeker:
            if target_index not in self.offset_dict.keys():
                for jump in range(40):
                    scan = None
                    insert_position = bisect.bisect_left(
                        self.seek_list, (target_index, 0)
                    )
                    if (
                        target_index < self.seek_list[0][0]
                        or target_index > self.seek_list[-1][0]
                    ):
                        raise Exception(
                            "Spectrum ID should be between"
                            " {0} and {1}".format(
                                self.seek_list[0][0], self.seek_list[-1][0]
                            )
                        )

                    element_before = self.seek_list[insert_position - 1]
                    spec_offset_m1 = target_index - element_before[0]

                    element_after = self.seek_list[insert_position]
                    spec_offset_p1 = element_after[0] - target_index

                    byte_diff_m1_p1 = element_after[1] - element_before[1]
                    scan_diff_m1_p1 = element_after[0] - element_before[0]

                    average_spec_between_m1_p1 = int(
                        round(byte_diff_m1_p1 / scan_diff_m1_p1)
                    )
                    # which side are we closer to ...
                    if spec_offset_m1 < spec_offset_p1:
                        # print("Closer to m1 - jumping forward")
                        jump_direction = "forwards"
                        jump_history["backwards"] = 0
                        jump_history["forwards"] += 1
                        byte_offset = element_before[1] + jump_history["forwards"] * (
                            offset_scale * average_spec_between_m1_p1 * spec_offset_m1
                        )
                        if (target_index - element_before[0]) < 10:
                            # quite close to target, stat at element before
                            # and read chunks until found
                            byte_offset = element_before[1]
                    else:
                        jump_direction = "backwards"
                        jump_history["forwards"] = 0
                        jump_history["backwards"] += 1
                        byte_offset = element_after[1] - jump_history["backwards"] * (
                            offset_scale * average_spec_between_m1_p1 * spec_offset_p1
                        )
                    byte_offset = int(byte_offset)
                    found_scan = False
                    chunk = b""
                    break_outer = False

                    for x in range(100):
                        seeker.seek(
                            max([os.SEEK_SET + byte_offset + x * chunk_size, 1])
                        )
                        chunk += seeker.read(chunk_size)
                    matches = re.finditer(regex_patterns.SPECTRUM_OPEN_PATTERN, chunk)
                    for _match_number, match in enumerate(matches):
                        if match is not None:
                            spec_info = match.groups()
                            spec_info = dict(zip(spec_info[0::2], spec_info[1::2]))
                            scan = int(re.search(b"[0-9]*$", spec_info[b"id"]).group())
                            # print(">>", _match_number, scan)
                            if jump_direction == "forwards":
                                if scan > target_index:
                                    # we went to far ...
                                    offset_scale = 0.1
                                    jump_history["forwards"] = 0
                                else:
                                    offset_scale = 1
                            if jump_direction == "backwards":
                                if scan < target_index:
                                    offset_scale = 0.1
                                    jump_history["backwards"] = 0
                                else:
                                    offset_scale = 1

                            if scan in self.offset_dict.keys():
                                continue
                            found_scan = True
                            new_entry = (
                                scan,
                                byte_offset + match.start(),
                            )
                            new_pos = bisect.bisect_left(self.seek_list, new_entry)
                            self.seek_list.insert(new_pos, new_entry)
                            self.offset_dict[scan] = (byte_offset + match.start(),)
                            if int(scan) == int(target_index):
                                # maybe jump from other boarder
                                break_outer = True
                                break
                    if break_outer:
                        break
                        if found_scan:
                            offset_scale = 1
                        else:
                            offset_scale += 1
                    if int(target_index) in self.offset_dict.keys():
                        break

            start = self.offset_dict[target_index]
            seeker.seek(start[0])
            match = None
            data = b""
            while b"</spectrum>" not in data:
                data += seeker.read(chunk_size)
            end = data.find(b"</spectrum>")
            seeker.seek(start[0])
            spec_string = seeker.read(end + len("</spectrum>"))
            spec_string = spec_string.decode("utf-8")
            spectrum = spec.Spectrum(XML(spec_string), measured_precision=5e-6)
            return spectrum

    def _build_index(self, from_scratch=False):
        """
        Build an index.

        A list of offsets to which a file pointer can seek
        directly to access a particular spectrum or chromatogram without
        parsing the entire file.

        Args:

            from_scratch(bool): Whether or not to force building the index from
                             scratch, by parsing the file, if no existing
                             index can be found.

        Returns:
            A file-like object used to access the indexed content by
            seeking to a particular offset for the file.
        """
        # Declare the pre-seeker
        seeker = self.get_binary_file_handler()
        self.offset_dict["TIC"] = None
        seeker.seek(0, 2)
        index_found = False

        spectrum_index_pattern = regex_patterns.SPECTRUM_INDEX_PATTERN
        for _ in range(1, 10):  # max 10kbyte
            # some converters fail in writing a correct index
            # we found
            # a) the offset is always the same (silent fail hurray!)
            sanity_check_set = set()
            try:
                seeker.seek(-1024 * _, 1)
            except:
                break
                # File is smaller than 10kbytes ...
            for line in seeker:
                match = regex_patterns.CHROMATOGRAM_OFFSET_PATTERN.search(line)
                if match:
                    self.offset_dict["TIC"] = int(bytes.decode(match.group("offset")))

                match_spec = spectrum_index_pattern.search(line)
                if match_spec is not None:
                    spec_byte_offset = int(bytes.decode(match_spec.group("offset")))
                    sanity_check_set.add(spec_byte_offset)

                match = regex_patterns.INDEX_LIST_OFFSET_PATTERN.search(line)
                if match:
                    index_found = True
                    index_list_offset = int(
                        match.group("indexListOffset").decode("utf-8")
                    )

            if index_found is True and self.offset_dict["TIC"] is not None:
                break

        if index_found is True:
            # Jumping to index list and slurpin all specOffsets
            seeker.seek(index_list_offset, 0)
            spectrum_index_pattern = regex_patterns.SPECTRUM_INDEX_PATTERN
            sim_index_pattern = regex_patterns.SIM_INDEX_PATTERN

            for line in seeker:
                match_spec = spectrum_index_pattern.search(line)
                if match_spec and match_spec.group("nativeID") == b"":
                    match_spec = None
                match_sim = sim_index_pattern.search(line)
                if self.index_regex is None:
                    if match_spec:
                        offset = int(bytes.decode(match_spec.group("offset")))
                        native_id = int(bytes.decode(match_spec.group("nativeID")))
                        self.offset_dict[native_id] = offset
                    elif match_sim:
                        offset = int(bytes.decode(match_sim.group("offset")))
                        native_id = bytes.decode(match_sim.group("nativeID"))
                        try:
                            native_id = int(
                                regex_patterns.SPECTRUM_ID_PATTERN2.search(
                                    native_id
                                ).group(2)
                            )
                        except AttributeError:
                            # match is None and has no attribute group,
                            # so use the whole string as ID
                            pass
                        self.offset_dict[native_id] = (offset,)
                else:
                    match = self.index_regex.search(line)
                    if match:
                        native_id = match.group("ID")
                        try:
                            native_id = int(native_id)
                        except ValueError:
                            pass
                        offset = match.group("offset")
                        self.offset_dict[native_id] = (offset,)

        elif from_scratch is True:
            seeker.seek(0)
            self._build_index_from_scratch(seeker)
        else:
            print("[Warning] Not index found and build_index_from_scratch is False")

        seeker.close()

    def _build_index_from_scratch(self, seeker):
        """Build an index of spectra/chromatogram data with offsets by parsing the file."""

        def get_data_indices(fh, chunksize=8192, lookback_size=100):
            """Get a dictionary with binary file indices of spectra and
            chromatograms in an mzML file.

            Will parse quickly through the file and find all occurences of
            <chromatogram ... id="..." and <spectrum ... id="..." using a
            regex.
            We dont use an XML parser here because we need to know the
            exact location of the filepointer which is usually not possible
            with common xml parsers.
            """
            chrom_positions = {}
            spec_positions = {}
            chromcnt = 0
            speccnt = 0
            # regexes to be used
            chromexp = re.compile(b'<\s*chromatogram[^>]*id="([^"]*)"')
            chromcntexp = re.compile(b'<\s*chromatogramList\s*count="([^"]*)"')
            specexp = re.compile(b'<\s*spectrum[^>]*id="([^"]*)"')
            speccntexp = re.compile(b'<\s*spectrumList\s*count="([^"]*)"')
            # go to start of file
            fh.seek(0)
            prev_chunk = ""
            while True:
                # read a chunk of data
                offset = fh.tell()
                chunk = fh.read(chunksize)
                if not chunk:
                    break

                # append a part of the previous chunk since we have cut in the middle
                # of the text (to make sure we dont miss anything, prev_chunk
                # is analyzed twice).
                if len(prev_chunk) > 0:
                    chunk = prev_chunk[-lookback_size:] + chunk
                    offset -= lookback_size
                prev_chunk = chunk

                # find all occurences of the expressions and add to the dictionary
                for m in chromexp.finditer(chunk):
                    chrom_positions[m.group(1).decode("utf-8")] = offset + m.start()
                for m in specexp.finditer(chunk):
                    spec_positions[m.group(1).decode("utf-8")] = offset + m.start()

                # also look for the total count of chromatograms and spectra
                # -> must be the same as the content of our dict!
                m = chromcntexp.search(chunk)
                if m is not None:
                    chromcnt = int(m.group(1))
                m = speccntexp.search(chunk)
                if m is not None:
                    speccnt = int(m.group(1))
            # Check if everything is ok (e.g. we found the right number of
            # chromatograms and spectra) and then return the dictionary.
            if chromcnt == len(chrom_positions) and speccnt == len(spec_positions):
                positions = {}
                positions.update(chrom_positions)
                positions.update(spec_positions)
            else:
                print(
                    "[ Warning ] Found {spec_count} spectra "
                    "and {chrom_count} chromatograms\n"
                    "[ Warning ] However Spectrum index list shows {speccnt} and "
                    "Chromatogram index list shows {chromcnt} entries".format(
                        spec_count=len(spec_positions),
                        chrom_count=len(chrom_positions),
                        speccnt=speccnt,
                        chromcnt=chromcnt,
                    )
                )
                print(
                    "[ Warning ] Updating offset dict with found offsets "
                    "but some might be still missing\n"
                    "[ Warning ] This may happen because your is file truncated"
                )
                positions = {}
                positions.update(chrom_positions)
                positions.update(spec_positions)
            return positions

        indices = get_data_indices(seeker)
        if indices is not None:
            tmp_dict = {}

            item_list = sorted(list(indices.items()), key=lambda x: x[1])
            for i in range(len(item_list)):
                key = item_list[i][0]
                tmp_dict[key] = (item_list[i][1],)

            self.offset_dict.update(tmp_dict)

            # make sure the list is sorted (for bisect)
            # self.info['offsetList'] = sorted(self.info['offsetList'])
            # self.info['seekable'] = True

        return

    def _interpol_search(self, target_index, chunk_size=8, fallback_cutoff=100):
        """
        Use linear interpolation search to find spectra faster.

        Arguments:
            target_index (str or int) : native id of the item to access

        Keyword Arguments:
            chunk_size (int)        : size of the chunk to read in one go in kb

        """
        # print('target ', target_index)
        seeker = self.get_binary_file_handler()
        seeker.seek(0, 2)
        chunk_size = chunk_size * 512
        lower_bound = 0
        upper_bound = seeker.tell()
        mid = int(upper_bound / 2)
        seeker.seek(mid, 0)
        current_position = seeker.tell()
        used_indices = set()
        spectrum_found = False
        spectrum = None
        while spectrum_found is False:
            jumper_scaling = 1
            file_pointer = seeker.tell()
            data = seeker.read(chunk_size)
            spec_start = self.spec_open.search(data)
            if spec_start is not None:
                spec_start_offset = file_pointer + spec_start.start()
                seeker.seek(spec_start_offset)
                spec_info = self.spec_open.search(data).groups()
                spec_info = dict(zip(spec_info[0::2], spec_info[1::2]))
                current_index = int(re.search(b"[0-9]*$", spec_info[b"id"]).group())

                self.offset_dict[current_index] = (spec_start_offset,)
                if current_index in used_indices:
                    if current_index > target_index:
                        jumper_scaling -= 0.1
                    else:
                        jumper_scaling += 0.1

                used_indices.add(current_index)

                dist = current_index - target_index
                if dist < -1 and dist > -(fallback_cutoff):
                    spectrum = self._search_linear(seeker, target_index)
                    spectrum_found = True
                    break
                elif dist > 0 and dist < fallback_cutoff:
                    while current_index > target_index:
                        offset = int(current_position - chunk_size)
                        seeker.seek(offset if offset > 0 else 0)
                        lower_bound = current_position
                        current_position = seeker.tell()
                        data = seeker.read(chunk_size)
                        if self.spec_open.search(data):
                            spec_info = self.spec_open.search(data).groups()
                            spec_info = dict(zip(spec_info[0::2], spec_info[1::2]))
                            current_index = int(
                                re.search(b"[0-9]*$", spec_info[b"id"]).group()
                            )
                    seeker.seek(current_position)
                    spectrum = self._search_linear(seeker, target_index)
                    spectrum_found = True
                    break

                if int(current_index) == target_index:

                    seeker.seek(spec_start_offset)
                    start, end = self._read_to_spec_end(seeker)
                    seeker.seek(start)
                    self.offset_dict[current_index] = (start, end)
                    xml_string = seeker.read(end - start)
                    spectrum = spec.Spectrum(XML(xml_string), measured_precision=5e-6)
                    spectrum_found = True
                    break

                elif int(current_index) > target_index:
                    scaling = target_index / current_index
                    seeker.seek(int(current_position * scaling * jumper_scaling))
                    upper_bound = current_position
                    current_position = seeker.tell()
                elif int(current_index) < target_index:
                    scaling = target_index / current_index
                    seeker.seek(int(current_position * scaling * jumper_scaling))
                    lower_bound = current_position
                    current_position = seeker.tell()

            elif len(data) == 0:
                sorted_int_keys = {
                    k: v for k, v in self.offset_dict.items() if isinstance(k, int)
                }
                sorted_keys = sorted(sorted_int_keys.keys())
                pos = (
                    bisect.bisect_left(sorted_int_keys, target_index) - 2
                )  # dat magic number :)
                try:
                    key = sorted_keys[pos]
                    spec_start_offset = self.offset_dict[key][0]
                except:
                    key = sorted_keys[pos]
                    spec_start_offset = self.offset_dict[key][0]
                seeker = self.get_binary_file_handler()
                seeker.seek(spec_start_offset)
                spectrum = self._search_linear(seeker, target_index)
                # seeker.close()
                spectrum_found = True
                break

        return spectrum

    def _read_to_spec_end(self, seeker, chunks_to_read=8):
        """
        Read from current seeker position to the end of the
        next spectrum tag and return start and end postition

        Args:
            seeker (_io.BufferedReader): Reader instance used in calling function

        Returns:
            positions (tuple): tuple with start and end postion of the spectrum
        """
        # start_pos = seeker.tell()
        chunk_size = 512 * chunks_to_read
        end_found = False
        start_pos = seeker.tell()
        data_chunk = seeker.read(chunk_size)
        while end_found is False:
            data_chunk += seeker.read(chunk_size)
            tag_end, seeker = self._read_until_tag_end(seeker)
            data_chunk += tag_end
            if regex_patterns.SPECTRUM_CLOSE_PATTERN.search(data_chunk):
                match = regex_patterns.SPECTRUM_CLOSE_PATTERN.search(data_chunk)
                end_pos = match.end()
                end_found = True
            elif regex_patterns.CHROMATOGRAM_CLOSE_PATTERN.search(data_chunk):
                match = regex_patterns.CHROMATOGRAM_CLOSE_PATTERN.search(data_chunk)
                end_pos = match.end()
                end_found = True
        return (start_pos, end_pos)

    def _read_extremes(self):
        """
        Read min and max spectrum ids. Required for binary jumps.

        Returns:
            seek_list (list): list of tuples containing spec_id and file_offset
        """
        chunk_size = 128000
        # chunk_size = 12800
        first_scan = None
        last_scan = None
        seek_list = []
        with open(self.path, "rb") as seeker:
            buffer = b""
            for x in range(100):
                try:
                    seeker.seek(os.SEEK_SET + x * chunk_size)
                except OSError:
                    break
                chunk = seeker.read(chunk_size)
                buffer += chunk
                match = regex_patterns.SPECTRUM_OPEN_PATTERN_SIMPLE.search(buffer)
                if match is not None:
                    id_match = regex_patterns.SPECTRUM_ID_PATTERN_SIMPLE.search(buffer)
                    try:
                        first_scan = int(
                            re.search(b"[0-9]*$", id_match.group("id")).group()
                        )
                    except ValueError:
                        first_scan = 0
                    #
                    seek_list.append(
                        (first_scan, seeker.tell() - chunk_size + match.start())
                    )
                    break
            buffer = b""
            seeker.seek(0, os.SEEK_END)
            for x in range(1, 100):
                try:
                    seeker.seek(-x * chunk_size, os.SEEK_END)
                except OSError:
                    break
                chunk = seeker.read(chunk_size)
                buffer = chunk + buffer
                # match = list(self.regex['spec_title_pattern'].finditer(buffer))

                matches = list(
                    regex_patterns.SPECTRUM_OPEN_PATTERN_SIMPLE.finditer(buffer)
                )
                if len(matches) != 0:
                    id_match = regex_patterns.SPECTRUM_ID_PATTERN_SIMPLE.search(
                        buffer[matches[-1].start() :]
                    )
                    last_scan = int(re.search(b"[0-9]*$", id_match.group("id")).group())
                    seek_list.append(
                        (last_scan, seeker.tell() - chunk_size + matches[-1].start())
                    )
                    break
        return seek_list

    def _search_linear(self, seeker, index, chunk_size=8):
        """
        Fallback to linear search if interpolated search fails.
        """
        data = None
        i = 0
        total_chunk_size = chunk_size * 512
        spec_start = None
        spec_end = None
        i = 0
        # print('target', index)
        while True:
            file_pointer = seeker.tell()

            data = seeker.read(total_chunk_size)
            string, seeker = self._read_until_tag_end(seeker)
            data += string

            spec_start = self.spec_open.search(data)
            if spec_start:
                spec_start_offset = file_pointer + spec_start.start()
                seeker.seek(spec_start_offset)
                spec_info = spec_start.groups()
                spec_info = dict(zip(spec_info[0::2], spec_info[1::2]))
                current_index = int(re.search(b"[0-9]*$", spec_info[b"id"]).group())
                # print(current_index)
                spec_end = self.spec_close.search(data[spec_start.start() :])
                if spec_end:
                    spec_end_offset = file_pointer + spec_end.end() + spec_start.start()
                    seeker.seek(spec_end_offset)
                while spec_end is None:

                    file_pointer = seeker.tell()

                    data = seeker.read(total_chunk_size)
                    string, seeker = self._read_until_tag_end(seeker)
                    data += string

                    spec_end = self.spec_close.search(data)
                    if spec_end:
                        spec_end_offset = file_pointer + spec_end.end()
                        self.offset_dict[current_index] = (
                            spec_start_offset,
                            spec_end_offset,
                        )
                        seeker.seek(spec_end_offset)
                        break

                if current_index == index:
                    seeker.seek(spec_start_offset)
                    spec_string = seeker.read(spec_end_offset - spec_start_offset)
                    self.offset_dict[current_index] = (
                        spec_start_offset,
                        spec_end_offset,
                    )
                    xml_string = XML(spec_string)
                    # seeker.close()
                    return spec.Spectrum(xml_string, measured_precision=5e-6)

    def _search_string_identifier(self, search_string, chunk_size=8):
        with self.get_binary_file_handler() as seeker:
            data = None
            total_chunk_size = chunk_size * 512
            spec_start = None

            # NOTE: This needs to go intp regex_patterns.py

            regex_string = re.compile(
                '<\s*spectrum[^>]*index="[0-9]+"\sid="({0})"\sdefaultArrayLength="[0-9]+">'.format(
                    "".join([".*", search_string, ".*"])
                ).encode()
            )

            search_string = search_string.encode()

            while True:
                file_pointer = seeker.tell()

                data = seeker.read(total_chunk_size)
                string, seeker = self._read_until_tag_end(seeker)
                data += string
                spec_start = regex_string.search(data)
                chrom_start = regex_patterns.CHROMO_OPEN_PATTERN.search(data)
                if spec_start:
                    spec_start_offset = file_pointer + spec_start.start()
                    current_index = spec_start.group(1)
                    if search_string in current_index:
                        seeker.seek(spec_start_offset)
                        start, end = self._read_to_spec_end(seeker)
                        seeker.seek(start)
                        spec_string = seeker.read(end)
                        xml_string = XML(spec_string)
                        return spec.Spectrum(xml_string, measured_precision=5e-6)
                elif chrom_start:
                    chrom_start_offset = file_pointer + chrom_start.start()
                    if search_string == chrom_start.group(1):
                        seeker.seek(chrom_start_offset)
                        start, end = self._read_to_spec_end(seeker)
                        seeker.seek(start)
                        chrom_string = seeker.read(end)
                        xml_string = XML(chrom_string)
                        return spec.Chromatogram(xml_string)
                elif len(data) == 0:
                    raise Exception("cant find specified string")

    def _read_until_tag_end(self, seeker, max_search_len=12):
        """
        Help make sure no splitted text appear in chunked data, so regex always find
        <spectrum ...>
        and
        </spectrum>
        """
        count = 0
        string = b""
        curr_byte = ""
        while (
            count < max_search_len
            and curr_byte != b">"
            and curr_byte != b"<"
            and curr_byte != b" "
        ):
            curr_byte = seeker.read(1)
            string += curr_byte
            count += 1
        return string, seeker

    def read(self, size=-1):
        """
        Read binary data from file handler.

        Keyword Arguments:
            size (int): Number of bytes to read from file, -1 to read to end of file

        Returns:
            data (str): byte string of len size of input data
        """
        return self.file_handler.read(size)

    def close(self):
        """ """
        self.file_handler.close()


if __name__ == "__main__":
    print(__doc__)



import sys
import os
import unittest

sys.path.append(os.path.abspath("."))

import pymzml
from pymzml.spec import PROTON
import pymzml.run as run
import test_file_paths
import numpy as np


class SpectrumMS2Test(unittest.TestCase):
    """
        BSA test file

    Peptide @
    Scan: 2548
    RT [min] 28.96722412109367
    Selected_precursor [(443.711242675781, 0.0)]

    """

    def setUp(self):
        """ """
        # self.paths = [
        #     os.path.join( DATA_FOLDER, file ) for file in DATA_FILES]
        self.paths = test_file_paths.paths
        path = self.paths[9]
        self.Run = run.Reader(path)
        self.spec = self.Run[2548]

    def test_scan_time(self):
        scan_time = self.spec.scan_time_in_minutes()
        scan_time2 = self.spec.scan_time_in_minutes()
        self.assertIsNotNone(scan_time)
        self.assertIsInstance(scan_time, float)
        self.assertEqual(round(scan_time, 4), round(28.96722412109367, 4))
        self.assertEqual(scan_time, scan_time2)

    def test_select_precursors(self):
        selected_precursor = self.spec.selected_precursors
        self.assertIsInstance(selected_precursor[0], dict)
        self.assertIsInstance(selected_precursor[0]["mz"], float)
        self.assertIsInstance(selected_precursor[0]["i"], float)
        self.assertIsInstance(selected_precursor[0]["charge"], int)
        assert selected_precursor[0]["mz"] == 443.711242675781
        assert selected_precursor[0]["i"] == 0.0
        assert selected_precursor[0]["precursor id"] is None

    def test_ion_mode(self):
        assert self.spec["positive scan"] is True

    def test_ion_mode_non_existent(self):
        assert self.spec["negative scan"] is None

    @unittest.skipIf(pymzml.spec.DECON_DEP is False, "ms_deisotope was not installed")
    def test_deconvolute_peaks(self):
        charge = 3
        test_mz = 430.313
        arr = np.array([(test_mz, 100), (test_mz + PROTON / charge, 49)])
        spec = self.Run[2548]
        spec.set_peaks(arr, "centroided")
        decon = spec.peaks("deconvoluted")
        self.assertEqual(len(decon), 1)
        decon_mz = (test_mz * charge) - charge * PROTON
        self.assertEqual(decon[0][0], decon_mz)
        self.assertEqual(decon[0][1], 149)  # 149 since itensities are 100 and 49
        self.assertEqual(decon[0][2], 3)

    def test_remove_precursor_peak(self):
        # breakpoint()
        # breakpoint()
        test_mz = 443.71124268  # precursor peak
        self.spec.set_peaks(np.array([(test_mz, 200)]), "centroided")
        self.spec.set_peaks(np.array([(test_mz, 200)]), "raw")
        assert self.spec.has_peak(test_mz)
        self.spec._transformed_mz_with_error = None
        new_peaks = self.spec.remove_precursor_peak()
        found_peaks = self.spec.has_peak(test_mz)
        assert len(found_peaks) == 0


if __name__ == "__main__":
    unittest.main(verbosity=3)



#!/usr/bin/env python
# -*- coding: latin-1 -*-
"""
Writer class for indexed gzipped files
"""

# Python mzML module - pymzml
# Copyright (C) 2010-2019 M. Kösters, C. Fufezan
#     The MIT License (MIT)

#     Permission is hereby granted, free of charge, to any person obtaining a copy
#     of this software and associated documentation files (the "Software"), to deal
#     in the Software without restriction, including without limitation the rights
#     to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#     copies of the Software, and to permit persons to whom the Software is
#     furnished to do so, subject to the following conditions:

#     The above copyright notice and this permission notice shall be included in all
#     copies or substantial portions of the Software.

#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#     IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#     FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#     AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#     LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#     OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#     SOFTWARE.

import struct
import time
import zlib
from collections import OrderedDict


class GSGW(object):
    """

    Generalized Gzip writer class with random access to indexed offsets.

    Keyword Arguments:
        file (string)        : Filename for the resulting file
        max_idx (int)        : max number of indices which can be saved in
                                this file
        max_idx_len (int)    : maximal length of the index in bytes, must
                                be between 1 and 255
        max_offset_len (int) : maximal length of the offset in bytes
        output_path (str)    : path to the output file

    """

    def __init__(
        self,
        file=None,
        max_idx=10000,
        max_idx_len=8,
        max_offset_len=8,
        output_path="./test.dat.igzip",
        comp_str=-1,
    ):
        self.Lock = False
        self._format_version = 1  # max 255!!!
        self.file_name = output_path
        self.max_idx_num = max_idx
        self.max_idx_len = max_idx_len
        self.max_offset_len = max_offset_len
        self.generic_header = OrderedDict(
            [
                ("MAGIC_BYTE_1", b"\x1f"),
                ("MAGIC_BYTE_2", b"\x8b"),
                ("COMPRESSION", b"\x08"),
                ("FLAGS", b"\x00"),
                ("DATE", b"\x00\x00\x00\x00"),
                ("XFL", b"\x02"),
                ("OS", b"\x03"),
            ]
        )
        self.index = OrderedDict()
        self.first_header_set = False
        self._file_out = None
        self._encoding = "latin-1"
        # magic bytes. FU+version
        self.index_magic_bytes = b"FU" + struct.pack("<B", self._format_version)
        self.crc32 = 0
        self.isize = 0
        self.comp_str = comp_str

    def __del__(self):
        """
        Close the file object properly after this object is deleted
        """
        self.file_out.close()

    def close(self):
        """
        Close the internal file object.
        """
        self.file_out.close()

    @property
    def file_out(self):
        """
        Output filehandler
        """
        if self._file_out is None:
            self._file_out = open(self.file_name, "wb")
        return self._file_out

    @property
    def encoding(self):
        """
        Returns the encoding used for this file
        """
        return self._encoding

    @encoding.setter
    def encoding(self, encoding):
        """
        Set the file encoding for the output file.
        """
        assert type(encoding) == str, "encoding must be a string"
        self._encoding = encoding

    def _write_gen_header(self, Index=False, FLAGS=None):
        """
        Write a valid gzip header with creation time, user defined flag fields
        and allocated index.

        Keyword Arguments:
            Index (bool)           : whether to or not to write an
                                        index into this header.
            FLAGS (list, optional) : list of flags (FTEXT, FHCRC, FEXTRA,
                                        FNAME) to set for this header.

        Returns:
            offset (int): byte offset of the file pointer
        """
        if FLAGS is None:
            FLAGS = []
        FTEXT, FHCRC, FEXTRA, FNAME = 1, 2, 4, 8  # extra field bit flags
        current_time = int(time.time())
        time_byte = struct.pack("<L", current_time)
        self.generic_header["DATE"] = time_byte
        if Index:
            self.generic_header["FLAGS"] = b"\x10"
        if FLAGS is not None:
            if "FTEXT" in FLAGS:
                self.generic_header["FLAGS"] = self.generic_header["FLAGS"] & FTEXT

            if "FHCRC" in FLAGS:
                header_crc32 = 0
                self.generic_header["FLAGS"] = self.generic_header["FLAGS"] & FHCRC
                for byte in self.generic_header.values():
                    header_crc32 = zlib.crc32(byte, header_crc32)

            if "FEXTRA" in FLAGS:
                self.generic_header["FLAGS"] = self.generic_header["FLAGS"] & FEXTRA

            if "FNAME" in FLAGS:
                self.generic_header["FLAGS"] = self.generic_header["FLAGS"] & FNAME

        for value in self.generic_header.values():
            self.file_out.write(value)
        if "FEXTRA" in FLAGS:
            # WRITE EXTRA FIELD
            pass

        if "FNAME" in FLAGS:
            # WRITE FNAME FIELD
            fName = self.file_name.split("/")[-1]

        if Index:
            self.generic_header["FLAGS"] = b"\x00"
            self.file_out.write(self.index_magic_bytes)
            self.file_out.write(struct.pack("<B", self.max_idx_len))
            self.file_out.write(struct.pack("<B", self.max_offset_len))
            self.index_offset = self.file_out.tell()
            self._allocate_index_bytes()

        if "FHCRC" in FLAGS:
            # WRITE checksum for header
            pass

        return self.file_out.tell()

    def _allocate_index_bytes(self):
        """
        Allocate 'self.max_index_num' bytes of length 'self.max_idx_len'
        in the header for inserting the index later on.
        """
        id_placeholder = self.max_idx_len * b"\x01"
        offset_placeholder = self.max_offset_len * b"\x01"
        for i in range(self.max_idx_num):
            self.file_out.write(id_placeholder)
            self.file_out.write(offset_placeholder)
        self.file_out.write(b"\x00")
        return

    def _write_data(self, data):
        """
        Write data into file-stream.

        Arguments:
            data (str): uncompressed data
        """
        Compressor = zlib.compressobj(
            self.comp_str, zlib.DEFLATED, -zlib.MAX_WBITS, zlib.DEF_MEM_LEVEL, 0
        )
        # compress data and flush (includes writing crc32 and isize)
        if isinstance(data, bytes) is False:
            data = bytes(data, "latin-1")
        self.crc32 = zlib.crc32(data)
        self.isize = len(data) % 2**32
        comp_data = Compressor.compress(data) + Compressor.flush()
        self.file_out.write(comp_data)
        self.file_out.write(struct.pack("<L", self.crc32))
        self.file_out.write(struct.pack("<L", self.isize))
        return

    def add_data(self, data, identifier):
        """
        Create a new gzip member with compressed 'data' indexed with 'index'.

        Arguments:
            data (str)         : uncompressed data to write to file
            index (str or int) : unique index for the data
        """
        if self.Lock is False:
            if len(self.index) + 1 > self.max_idx_num:
                print(
                    """
    WARNING: Reached maximum number of indexed data blocks
    '({0}), cannot add any more data!
                    """.format(
                        self.max_idx_num
                    )
                )
                return False

            if not self.first_header_set:
                self._write_gen_header(Index=True)
                self.first_header_set = True
            else:
                # do we need this?
                self._write_gen_header(Index=False)

            self.index[identifier] = self.file_out.tell()
            self._write_data(data)
            return
        else:
            raise Exception("Cant add any more data if index is already written")

    def _write_identifier(self, identifier):
        """
        Convert and write the identifier into output file.

        Arguments:
            identifier (str or int): identifier to write into index
        """
        id_format = "{0:\xAC>" + str(self.max_idx_len) + "}"
        identifier = str(identifier)
        identifier = id_format.format(identifier).encode("latin-1")
        self.file_out.write(identifier)
        return

    def _write_offset(self, offset):
        """
        Convert and write offset to output file.

        Arguments:
            offset (int): offset which will be formatted and written
                into file index
        """
        offset_format = "{0:\xAC>" + str(self.max_offset_len) + "}"
        offset = str(offset)
        offset = offset_format.format(offset).encode("latin-1")
        self.file_out.write(offset)
        return

    def write_index(self):
        """
        Only called after all the data is written, i.e. all calls to
        :func:`~GSGW.add_data` have been done.

        Seek back to the beginning of the file and write the index into the
        allocated comment bytes (see _write_gen_header(Index=True)).
        """
        self.Lock = True
        self.file_out.seek(self.index_offset)
        for identifier, offset in self.index.items():
            self._write_identifier(identifier)
            self._write_offset(offset)

    def __enter__(self):
        """
        Enable the with syntax for this class (entry point).
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Destructor when using this class with 'with .. as'."""
        self.file_out.close()


if __name__ == "__main__":
    print(__doc__)



#!/usr/bin/env python
# -*- coding: utf-8 -*
"""
MSNumpress decoder class


Authors:

Manuel Kösters
Christian Fufezan
"""
import sys
import struct
import math

import numpy as np


class MSNumpress:
    """
    The library provides implementations of 4 different algorithms,
    1 designed to compress first order smooth data like retention time or M/Z
    arrays, 1 designed to transform first order smooth data for more efficient
    zlib compression, and 2 for compressing non-smooth data with lower
    requirements on precision like ion count arrays.


    ..Note::
        This is a Python implementation of the Golomb-Rice encoding,
        also known as MS-Numpress. As such it is considerably slower than
        the C implementation, which we wrapped into cython. Please make
        yourself a favor and use the the wrapper/c implementation and only
        as a last result this Python implementation.

    """

    def __init__(self, data=None):
        """
        Initialize stateful Numpress De- and Encoding.

        Args:
            array (list): data array with compressed or decompressed data
        """
        self.is_little_endian = True if sys.byteorder == "little" else False
        self.Filler = {
            8: "00000000",
            7: "0000000",
            6: "000000",
            5: "00000",
            4: "0000",
            3: "000",
            2: "00",
            1: "0",
            0: "",
            9: "f",
            10: "ff",
            11: "fff",
            12: "ffff",
            13: "fffff",
            14: "ffffff",
            15: "fffffff",
        }
        # call check what data was set ...
        self._encoded_data = None
        self._decoded_data = None

        if type(data) == bytearray:
            # set data is encoded
            self.data_state = "encoded"
            self._encoded_data = data
        elif type(data) == list:
            # set data is decoded
            self.data_state = "decoded"
            self._decoded_data = data
        else:
            pass

    @property
    def decoded_data(self):
        if self._decoded_data is None:
            raise Exception("decoded data is not set")
        return self._decoded_data

    @decoded_data.setter
    def decoded_data(self, data):
        if type(data) is not list and type(data) is not np.ndarray:
            raise Exception("data must be list")
        self._decoded_data = data

    @property
    def encoded_data(self):
        if self._encoded_data is None:
            raise Exception("encoded data is not set")
        return self._encoded_data

    @encoded_data.setter
    def encoded_data(self, data):
        """Set the data."""
        if type(data) is not bytearray:
            raise Exception("data must be bytearray")
        self._encoded_data = data

    def _linear_fixed_point(self):
        """
        Calculate the optimal predictor coefficient for linear prediction model
        for the Numpress Linear algorithm.

        Returns:
            fixed_point(int): scaling factor for linear prediction
        """
        data = self.decoded_data
        data_size = len(data)
        if data_size == 0:
            return 0
        if data_size == 1:
            return math.floor(0xFFFFFFFF) / data[0]
        max_double = max(data[:2])
        for i in range(2, data_size):
            extrapol = data[i - 1] + (data[i - 1] - data[i - 2])
            diff = data[i] - extrapol
            max_double = max([max_double, math.ceil(abs(diff) + 1)])
        return math.floor(0x7FFFFFFF / max_double)

    def _slof_fixed_point(self):
        """
        Calculate the optimal predictor coefficient for linear prediction model
        for the Numpress Slof algorithm.

        Returns:
            fixed_point(int): scaling factor for linear prediction
        """
        data = self.decoded_data
        if len(data) == 0:
            return 0
        max_double = 1
        for i in range(len(data)):
            x = math.log(data[i] + 1)
            max_double = max([max_double, x])
        fp = math.floor(0xFFFF / max_double)
        return fp

    def _encodeInt(self, integer):
        """
        # RENAME FUNCTION :)
        Encode a given 4 byte integer by truncating leading 1s and 0s
        and saving them in a halfbyte along with the original data

        Args:
            integer (int): Value to be compressed

        Returns:
            bytes (bytearray): encoded bytes
        """
        mask = 0xF0000000

        try:
            integer_as_bytes = struct.pack(">i", integer)
        except:
            integer_as_bytes = None
        init = integer & 0xF0000000
        results = bytearray()

        if integer_as_bytes == b"\x00\x00\x00\x00":
            results.append(0x8)
            x = 8

        elif init == 0:
            x = 0
            for b in integer_as_bytes:
                first_half_byte = (0xFF & b) >> 4
                second_half_byte = 0xF & b
                if first_half_byte == 0:
                    x += 1
                    if second_half_byte == 0:
                        x += 1
                        continue
                    else:
                        results.append(x)
                        break
                else:
                    results.append(x)
                    break

        else:
            for x in range(8):
                m = mask >> 4 * x
                if integer & m == m:
                    continue
                else:
                    results.append(x + 8)
                    break
            if len(results) == 0:
                results.append(x + 8)

        if x < 8:
            for m in range(8 - x):
                r_to_l_mask = 0xF
                np_byte = (integer >> 4 * m) & r_to_l_mask
                results.append(np_byte)

        return results

    def _decodeInt(self, encodedInt):
        """
        Decode an integer from bytes encoded by :py:func:`_encodeInt`

        Args:
            encodedInt (bytearray): encoded bytes

        Returns:
            result (int): decoded form of the encoded integer
        """
        hex_string = ""
        fill = "0"
        for pos, byte in enumerate(encodedInt):
            if pos == 0:
                leading = byte & 0xF
                if leading > 8:
                    leading -= 8
                    fill = "f"
            else:
                hex_string = str(hex(byte))[2:] + hex_string
        hex_string = fill * leading + hex_string
        x = int(hex_string, 16)
        if x > 0x7FFFFFFF:
            x -= 0x100000000

        return x

    def encode_linear(self):
        """
        Use Linear prediction and Golomb coding to compress an array
        of mz values.

        Returns:
            result (bytearray): Golomb encoded bytearray

        EXPAMPLE

        IN:
            [ 12.23123, 14.21431, 23.22222 ]
        OUT:
            (b'\x50\x0c\x0f\x08\x51\x02\x03\x01\x23\x00') <class 'bytearray'>
        """

        ints = [0 for x in range(3)]
        data = self.decoded_data
        fp = self._linear_fixed_point()

        encoded_fixed_point = self._encode_fixed_point(fp)
        result = bytearray(encoded_fixed_point)
        if len(data) == 0:
            return 8

        ints[1] = int(round(data[0] * fp))

        for i in range(4):
            result.append((ints[1] >> (i * 8)) & 0xFF)

        if len(data) == 1:
            return 12

        ints[2] = int(round(data[1] * fp))

        for i in range(4):
            result.append((ints[2] >> (i * 8)) & 0xFF)

        enc_diff = bytearray()
        for i in range(2, len(data)):
            ints[0] = ints[1]
            ints[1] = ints[2]
            ints[2] = int(round(data[i] * fp))
            extrapol = ints[1] + (ints[1] - ints[0])

            diff = ints[2] - extrapol
            enc_diff += self._encodeInt(diff)
            for i in range(1, len(enc_diff), 2):
                final_byte = enc_diff[i] & 0xF | enc_diff[i - 1] << 4
                result.append(final_byte)
            if len(enc_diff) % 2 != 0:
                enc_diff = bytearray([enc_diff[-1]])
            else:
                enc_diff = bytearray()

        self.encoded_data = result
        return result

    def decode_linear(self):
        """
        Decode a Golomb encoded bytearray to its corresponding numpy array.

        Returns:
            result (np.ndarray): NumLin decoded numpy array of the
                original data

        """
        ints = [0 for x in range(3)]
        data = self.encoded_data

        if len(data) < 8:
            raise Exception("Corrupt input data.\nnot enough bytes to read fixed point")

        enc_fp = data
        fp = self._decode_fixed_point(enc_fp)

        if len(data) < 12:
            raise Exception("Corrupt input data.\nnot enough bytes to read first value")

        for i in range(4):
            ints[1] = ints[1] | ((0xFF & (data[8 + i])) << (i * 8))
        i1 = ints[1] / fp

        if len(data) < 16:
            raise Exception("Corrupt input data\nnot enough bytes to read second value")

        for i in range(4):
            ints[2] = ints[2] | ((0xFF & (data[12 + i])) << (i * 8))
        i2 = ints[2] / fp

        diff_vals = self._decode_ints_from_bytearray(16, fp)
        result = [0 for x in range(len(diff_vals) + 2)]
        result[0] = i1
        result[1] = i2
        i = 2
        for diff in diff_vals:
            ints[0] = ints[1]
            ints[1] = ints[2]
            ints[2] = diff
            extrapol = ints[1] + (ints[1] - ints[0])
            val = extrapol + ints[2]
            result[i] = val / fp
            ints[2] = val
            i += 1

        return result

    def _decode_ints_from_bytearray(self, pos, normalization=1):
        """
        Decode truncated integer encoded values and end position
        in bytearray.

        Args:
            pos (int) : start position in array

        Returns:
            results (list): ...
        """
        # results = np.empty(len(self.encoded_data) * 2)
        results = [0 for x in range(len(self.encoded_data) * 2)]
        leading_count = None
        i = 0
        while pos < len(self.encoded_data):
            current_byte = "{0:02x}".format(self.encoded_data[pos])
            for byte_pos in [0, 1]:
                hb = current_byte[byte_pos]
                if leading_count is None:
                    if hb == "8":
                        results[i] = 0
                        i += 1
                    else:
                        leading_count = int(hb, 16)
                        hex_str = ""
                        cb = leading_count - (8 * math.floor(leading_count / 8))
                else:
                    hex_str = "{0}{1}".format(hb, hex_str)
                    cb += 1
                    try:
                        1 / (cb - 8)
                    except:
                        final_int = int(self.Filler[leading_count] + hex_str, 16)
                        if final_int > 0x7FFFFFFF:
                            final_int -= 0x100000000
                        results[i] = final_int
                        i += 1
                        leading_count = None
            pos += 1
        return results[:i]

    def encode_pic(self):
        """
        Encode Ion count data by rounding to the next integer and
        store in a truncated form.

        Returns:
            result (bytearray): array with rounded
                ion count data in bytes
        """
        # need to be optimized
        res = bytearray()
        final = bytearray()

        for val in self.decoded_data:
            val = int(round(val))  # simply round
            enc_int = self._encodeInt(val)
            res.extend(enc_int)

        # assure res has an even length
        if len(res) % 2 != 0:
            res.append(0)

        # pull halfbytes together
        for i in range(1, len(res), 2):
            val = (res[i - 1] << 4) | res[i] & 0xF
            final.append(val)

        self.encoded_data = final
        return final

    def decode_pic(self):
        """
        Decode ion count data compressed with :py:func:`encodePic`

        Returns:
            result (array): decoded Ion count data as numpy array
        """
        results = []
        data = self.encoded_data
        read_first = True
        hb_to_read = 0
        current_value = bytearray()

        # TODO fix problem when 2 or more zeroes appear after each other
        # => if first hbc is 8, the next number is a hbc again and not data

        for val in data:
            if hb_to_read == 0:
                if read_first is True:
                    current_value = bytearray()
                    count = (0xFF & val) >> 4
                    current_value.append(count)
                    hb_to_read = 8 - count
                    read_first = False
                else:
                    current_value = bytearray()
                    count = 0xF & val
                    current_value.append(count)
                    hb_to_read = 8 - count
                    read_first = True
            if hb_to_read != 0:
                if hb_to_read == 1:
                    if read_first is True:
                        hb = (0xFF & val) >> 4
                        current_value.append(hb)
                        hb_to_read -= 1
                        dec_int = self._decodeInt(current_value)
                        current_value = bytearray()
                        results.append(dec_int)
                        count = 0xF & val
                        current_value.append(count)
                        hb_to_read = 8 - count
                        continue
                    else:
                        hb = 0xF & val
                        current_value.append(hb)
                        hb_to_read -= 1
                        dec_int = self._decodeInt(current_value)
                        current_value = bytearray()
                        results.append(dec_int)
                        read_first = True
                        continue
                if read_first is True:
                    hb1 = (0xFF & val) >> 4
                    hb2 = 0xF & val
                    current_value.append(hb1)
                    current_value.append(hb2)
                    hb_to_read -= 2
                    if hb_to_read == 0:
                        dec_int = self._decodeInt(current_value)
                        results.append(dec_int)
                        current_value = bytearray()
                        continue
                else:
                    hb = 0xF & val
                    current_value.append(hb)
                    read_first = True
                    hb_to_read -= 1

        self.decoded_data = results

        return results

    def encode_slof(self):
        """
        encode ion count data by multiplying the logarithm of each value with
        a scaling factor, rounding to nearest integer, and saving the 2 LSB.

        Returns:
            result (bytearray): short logged float compressed bytearray
        """
        data = self.decoded_data
        self.fixed_point = self._slof_fixed_point()
        res = self._encode_fixed_point(self.fixed_point)
        for i in range(len(data)):
            x = math.floor((math.log(data[i] + 1)) * self.fixed_point + 0.5)
            res.append(0xFF & x)
            res.append(x >> 8)
        self.encoded_data = res
        return res

    def decode_slof(self):
        """
        Decode short logged float compressed data.

        Returns:
            result (bytearray): array with decoded ion count data
        """
        data = self.encoded_data
        if len(data) < 8:
            return -1
        res = list()
        fixed_point = self._decode_fixed_point(data)

        for i in range(8, len(data), 2):
            x = 0xFF & data[i] | ((0xFF & data[i + 1]) << 8)
            res.append(math.exp((0xFFFF & x) / fixed_point) - 1)

        self.decoded_data = res
        return res

    def encode_safe(self):
        """
        Transform values by linear prediction without scaling or truncated
        representation.

        Returns:
            result (bytearray): transformed data in bytes
        """
        # result = bytearray()
        # assert data format is correct
        # save_first_2_values_unencoded()
        # for val in self.data[2:]:
        #   pred = do_linear_prediction()
        #   enc  = do_Golomb_encoding()
        #   result.extend(enc)
        # return result
        pass

    def decode_safe(self):
        # if not self.data_state == 'decoded':
        #   raise Exception('No encoded data found\nPlease set encoded Data')

        # result = np.ndarray()
        # int1 = read_int1(8)
        # int2 = read_int2(8)
        # add ints to result
        # for x in range(0, len(self.data, 2)):
        #   enc_int = self.data[x:x+8]
        #   result.append(dec_int)
        pass

    def _encode_fixed_point(self, value):
        """
        Save a given float as bytes.

        Args:
            value(float): fixed point
        Returns:
            res (bytearray): encoded fixed point
        """
        res = bytearray(8)
        fp = value
        fp = struct.pack("<d", fp)
        fp = struct.unpack("<q", fp)[0]
        if self.is_little_endian:
            for i in range(8):
                res[7 - i] = (fp >> (8 * i)) & 0xFF
        else:
            for i in range(7, -1, -1):
                res[7 - i] = (fp >> (8 * i)) & 0xFF
        return res

    def _decode_fixed_point(self, value):
        """
        Decode a given bytearray to float.

        Args:
            value(bytearray): encoded fixed point
        Returns:
            res (float): encoded fixed point
        """
        fp = 0
        if self.is_little_endian:
            for i in range(8):
                fp = fp | ((0x00FF & value[7 - i]) << (8 * i))
        else:
            for i in range(7, -1, -1):
                fp = fp | ((0x00FF & value[7 - i]) << (8 * i))
        fp = struct.pack("<q", fp)
        fp = struct.unpack("<d", fp)[0]
        return fp


if __name__ == "__main__":
    print(__doc__)



#!/usr/bin/env python3

import os

import pymzml


def main():
    """
    Testscript to highlight the function name changes in the Spectrum class.

    Note:
        Please adjust any old scripts to the new syntax.

    usage:

        ./deprecation_check.py

    """

    example_file = os.path.join(
        os.path.dirname(__file__), os.pardir, "tests", "data", "example.mzML"
    )
    run = pymzml.run.Reader(example_file)
    spectrum_list = []
    for pos, spectrum in enumerate(run):
        spectrum_list.append(spectrum)
        spectrum.hasPeak((813.19073486))
        spectrum.extremeValues("mz")
        spectrum.hasOverlappingPeak(813.19073486)
        spectrum.highestPeaks(1)
        spectrum.estimatedNoiseLevel()
        spectrum.removeNoise()
        spectrum.transformMZ(813.19073486)
        if pos == 1:
            spectrum.similarityTo(spectrum_list[0])
            break


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Part of pymzml test cases
"""
# import sys
# import os
# # import PyNumpress
# import pymzml
# import pymzml.decoder as decoder
# import time
# import unittest
# import numpy as np
# # import PyNumpress as pnp
# import zlib
# from base64 import b64encode as b64enc
# import test_file_paths


# class DecoderTest(unittest.TestCase):

#     def assertPeaksIdentical(self, peaks1, peaks2, msg=None):
#         self.assertEqual(len(peaks1), len(peaks2))#, msg='List have different number of peaks!')
#         for x in range(len(peaks1)):
#             self.assertCountEqual(peaks1[x], peaks2[x], msg=msg)

#     def setUp(self):
#         self.paths = test_file_paths.paths
#         self.Decoder = pymzml.Decoder
#         self.Run  = pymzml.run.Reader(self.paths[2])

#     def test_decode_numpress(self):
#         arr = np.asarray([1,2,3], dtype=np.float64)
#         dec = pnp.MSNumpress([])
#         enc_np = dec.encode_linear(arr, dec.optimal_linear_fixed_point(arr))
#         enc_np_zlib = zlib.compress(enc_np)
#         enc_np_zlib_b64 = b64enc(enc_np_zlib)
#         comp = ['zlib', 'MS-Numpress linear prediction compression']
#         d_type, decoded = decoder._decode(enc_np_zlib_b64, comp, 3, '32-bit float', 'i')
#         self.assertCountEqual(arr, decoded)

#     def test_decode_numpress(self):
#         test_array = np.asarray([1,2,3], dtype=np.float64)
#         MSNumpress = PyNumpress.MSNumpress([])
#         nump_enc   = MSNumpress.encode_linear(test_array, MSNumpress.optimal_linear_fixed_point(test_array))
#         zlib_nump_enc = zlib.compress(nump_enc)
#         b64_zlib_nump_enc =b64enc( zlib_nump_enc )
#         d_type, arr = decoder._decode(b64_zlib_nump_enc, ['zlib', 'ms-np-slof'], 3, '64-bit float', 'i')
#         self.assertIsNotNone(len(arr))
#         self.assertIsInstance(arr, list)


#     def test_decode_32_bit_no_compression(self):
#         b64_array = 'cgDIQjgByEL+AchCxQLIQiGL7kIjjO5CJY3uQieO7kIpj+5CK5DuQi2R7kIvku5CMpPuQjeU7kI5le5CO5buQj2X7kK6QgJDTUMCQ+FDAkN0RAJDB0UCQ5pFAkMuRgJDwUYCQ1RHAkPoRwJDe0gCQw9JAkOiSQJDj/YGQyr3BkPF9wZDYfgGQ/z4BkOX+QZDM/oGQ876BkNp+wZDBfwGQ6L8BkM9/QZD2P0GQ3T+BkPmBRVDmgYVQ04HFUMCCBVDtggVQ2oJFUMfChVD0woVQ4cLFUM7DBVD8AwVQ6QNFUNYDhVDDA8VQ8EPFUN1EBVDKREVQ/UFFkOrBhZDYQcWQxcIFkPNCBZDgwkWQzkKFkPvChZDpQsWQ1sMFkMRDRZDxw0WQ34OFkM0DxZDw6gZQ4CpGUM9qhlD+aoZQ7arGUNzrBlDL60ZQ+ytGUOprhlDZq8ZQyKwGUPfsBlDnLEZQ0iiGkMGoxpDxaMaQ4OkGkNCpRpDAKYaQ7+mGkN9pxpDPKgaQ/+oGkO9qRpDfKoaQzqrGkO0o0dDy6RHQ+OlR0P6pkdDEahHQympR0NAqkdDWKtHQ2+sR0OHrUdDnq5HQ7avR0PNsEdDaT1RQ5U+UUPAP1FD7EBRQxhCUUNEQ1FDcERRQ5xFUUPHRlFD9EdRQyBJUUNLSlFDd0tRQ3p0bEPidWxDS3dsQ7N4bEMbemxDg3tsQ+t8bENTfmxDvH9sQySBbEONgmxD9YNsQ12FbEPGhmxDDn6EQ+R+hEO5f4RDj4CEQ2WBhEM6goRDEIOEQ+aDhEO7hIRDkYWEQ2eGhEM8h4RDEoiEQ5x6hUN0e4VDTHyFQyR9hUP9fYVD1X6FQ61/hUOFgIVDXYGFQzWChUMNg4VD5YOFQ72EhUOVhYVDlnaGQ3B3hkNLeIZDJXmGQwB6hkPaeoZDtHuGQ498hkNpfYZDRH6GQx5/hkP5f4ZD04CGQ66BhkOIgoZDY4OGQz2EhkMM9oZD6PaGQ8T3hkOf+IZDe/mGQ1f6hkMy+4ZDDvyGQ+r8hkPF/YZDof6GQ33/hkNYAIdDNAGHQxACh0PrAodDG9CIQ/vQiEPb0YhDu9KIQ5vTiEN71IhDXNWIQzzWiEMc14hD/NeIQ9zYiEO82YhDnNqIQ33biENvQYxDV0KMQ0BDjEMpRIxDEUWMQ/pFjEPjRoxDy0eMQ7RIjEOdSYxDhkqMQ25LjENXTIxDT3+MQzmAjEMigYxDC4KMQ/WCjEPeg4xDx4SMQ7CFjEOahoxDg4eMQ2yIjENWiYxDP4qMQyiLjEMSjIxD+4yMQ+WNjEMhAI1DCwGNQ/YBjUPgAo1DywONQ7UEjUOgBY1DigaNQ3UHjUNgCI1DSgmNQzUKjUMfC41DCgyNQ/UMjUP0fY1D4H6NQ8x/jUO3gI1Do4GNQ4+CjUN7g41DZ4SNQ1KFjUM+ho1DKoeNQxaIjUMCiY1D7YmNQ9mKjUO6eY5DqXqOQ5d7jkOFfI5Dc32OQ2F+jkNQf45DPoCOQyyBjkMago5DCYOOQ/eDjkPlhI5D1IWOQ8KGjkOwh45Dn4iOQ3n6jkNo+45DWPyOQ0f9jkM3/o5DJv+OQxYAj0MFAY9D9QGPQ+QCj0PUA49DwwSPQ7IFj0OiBo9DkQePQ4EIj0P7f5VD+4CVQ/uBlUP7gpVD+4OVQ/uElUP7hZVD+4aVQ/uHlUP7iJVD/ImVQ/yKlUP8i5VD/IyVQ/yNlUP8jpVD/I+VQ4AtmkOMLppDmC+aQ6UwmkOxMZpDvTKaQ8kzmkPVNJpD4jWaQ+42mkP6N5pDBjmaQxM6mkMfO5pDbF3DQ+pew0NpYMND52HDQw=='
#         d_type, arr = decoder._decode(b64_array, 'no compression', 343, '32-bit float', 'i')
#         self.assertIsNotNone(len(arr))
#         self.assertIsInstance(arr, np.ndarray)

#     def test_pool_decode_TIC(self):
#         """
#         """
#         spec  = self.Run["TIC"]
#         spec2 = self.Run["TIC"]
#         paramsMZ = spec._get_encoding_parameters('time array')
#         paramsMZ += ('time',)
#         paramsI = spec._get_encoding_parameters('intensity array')
#         paramsI += ('i',)

#         s = time.time()
#         self.Decoder.pool_decode([paramsMZ, paramsI], spec._register)
#         t1 = time.time() - s
#         peaks1 = spec.profile


#         s = time.time()
#         peaks2 = list(zip(spec.time, spec.i))
#         t2 = time.time() - s
#         self.assertPeaksIdentical(peaks1, peaks2)

#         assert t1 < t2, 'parallel version is slower than normal version:\n' \
#                 'Parallel took {0:.5f} seconds\n' \
#                 'Normal took {1:.05f} seconds\n'.format(t1, t2)

#     def test_pool_decode_spec4000(self):
#         spec3 = self.Run[5]
#         spec4 = self.Run[5]
#         paramsMZ = spec3._get_encoding_parameters('m/z array')
#         paramsMZ += ('mz',)
#         paramsI = spec3._get_encoding_parameters('intensity array')
#         paramsI += ('i',)

#         s = time.time()
#         self.Decoder.pool_decode([paramsMZ, paramsI], spec3._register)
#         t1 = time.time() - s
#         peaks3 = spec3.peaks

#         s = time.time()
#         peaks4 = list(zip(spec4.mz, spec4.i))
#         t2 = time.time() - s
#         self.assertPeaksIdentical(peaks3, peaks4)

#         assert t1 < t2, 'parallel version is slower than normal version:\n' \
#                 'Parallel took {0:.5f} seconds\n' \
#                 'Normal took {1:.05f} seconds\n'.format(t1, t2)

#     def test_big_parallel_array(self):
#         arr = np.asarray([x for x in range(1000000)], dtype=np.float64)
#         dec = pnp.MSNumpress([])
#         enc_np = dec.encode_linear(arr, dec.optimal_linear_fixed_point(arr))
#         enc_np_zlib = zlib.compress(enc_np)
#         enc_np_zlib_b64 = b64enc(enc_np_zlib)
#         comp = ['zlib', 'MS-Numpress linear prediction compression']
#         params = [
#             (enc_np_zlib_b64, comp, 100000, '32-bit float', 'i'),
#             (enc_np_zlib_b64, comp, 100000, '32-bit float', 'mz')
#         ]
#         s = time.time()
#         self.Decoder.pool_decode(params, lambda x,y :print(x,y))
#         t1 = time.time() - s

#         s = time.time()
#         d_type, decoded = decoder._decode(*params[0])
#         d_type, decoded = decoder._decode(*params[1])
#         t2 = time.time() - s
#         print(t1, t2)
#         assert t1 < t2

if __name__ == "__main__":
    unittest.main(verbosity=3)



import sys
import os

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

import pymzml.run as run

try:
    import numpy as np
except:
    np = None
import unittest
import test_file_paths


class ChromatogramTest(unittest.TestCase):
    def assertPeaksIdentical(self, peaks1, peaks2, msg=None):
        self.assertEqual(
            len(peaks1), len(peaks2)
        )  # , msg='List have different number of peaks!')
        for x in range(len(peaks1)):
            self.assertCountEqual(peaks1[x], peaks2[x], msg=msg)

    def setUp(self):
        self.paths = test_file_paths.paths
        path = self.paths[2]
        self.Run_np = run.Reader(path)
        self.chrom = self.Run_np["TIC"]

    def test_time(self):
        time = self.chrom.time
        mz = self.chrom.mz
        self.assertCountEqual(time, mz)
        intensity = self.chrom.i

    def test_i(self):
        self.chrom.profile = [(1, 10), (2, 20), (3, 30)]
        peaks = self.chrom.peaks()
        print(self.chrom.peaks())
        self.assertPeaksIdentical(peaks, [(1, 10), (2, 20), (3, 30)])

    def test_profile(self):
        profile = self.chrom.profile
        self.assertIsNotNone(len(profile))
        if np:
            self.assertIsInstance(profile, np.ndarray)
        else:
            self.assertIsInstance(profile, list)


if __name__ == "__main__":
    unittest.main(verbosity=3)



import unittest
import struct
import math

try:
    import numpy as np
except:
    np = None

from pymzml.ms_numpress import MSNumpress


@unittest.skipIf(np is None, "Numpy is required for this test.")
class test_MSNumpress(unittest.TestCase):

    """
    unittest for MSNumpress en- and decoding
    """

    def setUp(self):
        """ """
        self.mz_data = np.asarray([100.1, 100.01, 100.001, 100.0001], dtype=np.float64)
        self.i_data = [5e6, 6.5e5, 2e6, 12e6]
        self.i_slof_data = [1, 2, 4]
        self.integer = 23
        # self._encoded_int = bytes([ 0x06, 0x07, 0x01 ])
        self.Decoder = MSNumpress()
        self.fixed_point = 10000

    def test_enc_int_leading_zeros(self):
        encoded_integer = self.Decoder._encodeInt(1)
        self.assertEqual(encoded_integer[0], 0x07)
        encoded_integer = self.Decoder._encodeInt(16)
        self.assertEqual(encoded_integer[0], 0x06)

    def test_enc_int_leading_ones(self):
        bytearray(b"\x08\x0f\x0f\x0f\x0f\x0f\x0f\x0f\x07")
        encoded_integer = self.Decoder._encodeInt((2**31) - 1)
        self.assertEqual(encoded_integer[0], 0x08)
        self.assertEqual(encoded_integer[1], 0x0F)

    def test_encodeInt_negative(self):
        """ """
        negative_integer = -1
        encoded_integer = self.Decoder._encodeInt(negative_integer)
        # print('encoded', encoded_integer)
        self.assertEqual(encoded_integer[0], 0xF)
        self.assertEqual(encoded_integer[1], 0xF)

    def test_encodeInt_0(self):
        encoded_integer = self.Decoder._encodeInt(0)
        self.assertEqual(len(encoded_integer), 1, msg="{}".format(encoded_integer))
        self.assertEqual(encoded_integer[0], 0x08)

    def test_encodeInt_256(self):
        encoded_integer = self.Decoder._encodeInt(256)
        self.assertEqual(len(encoded_integer), 4)
        self.assertEqual(encoded_integer[0], 0x05)
        self.assertEqual(encoded_integer[1], 0x00)
        self.assertEqual(encoded_integer[2], 0x00)
        self.assertEqual(encoded_integer[3], 0x01)

    def test_encode_4096(self):
        encoded_integer = self.Decoder._encodeInt(4096)
        self.assertEqual(len(encoded_integer), 5)
        self.assertEqual(encoded_integer[0], 0x04)
        self.assertEqual(encoded_integer[1], 0x00)
        self.assertEqual(encoded_integer[2], 0x00)
        self.assertEqual(encoded_integer[3], 0x00)
        self.assertEqual(encoded_integer[4], 0x01)

    def test_dec_int_leading_zeros(self):
        encoded_integer = bytearray([0x07, 0x01])
        decoded_integer = self.Decoder._decodeInt(encoded_integer)
        self.assertEqual(decoded_integer, 1)

    def test_encode_decode_int(self):
        encoded_integer = self.Decoder._encodeInt(200)
        # print(encoded_integer)
        decoded_integer = self.Decoder._decodeInt(encoded_integer)
        self.assertEqual(decoded_integer, 200)

    def test_dec_int_leading_ones(self):

        encoded_integer = bytearray(b"\x08\n\x03\x04\x01\x05\x03\x07\x07")
        decoded_integer = self.Decoder._decodeInt(encoded_integer)
        # print(decoded_integer)
        self.assertEqual(
            decoded_integer,
            1999967290,
            msg="result: {}\nexpected: {}".format(decoded_integer, hex(1999967290)),
        )

    def test_dec_int_negative(self):
        encoded_negative_int = bytearray([0x0F, 0x0F])
        decoded_negative_int = self.Decoder._decodeInt(encoded_negative_int)
        self.assertEqual(decoded_negative_int, -1)

    def test_encode_slof(self):
        self.Decoder.decoded_data = self.i_slof_data
        encoded_array = self.Decoder.encode_slof()

        self.assertEqual(len(encoded_array), 14)

        self.assertEqual(encoded_array[0], 0x40)
        self.assertEqual(encoded_array[1], 0xE3)
        self.assertEqual(encoded_array[2], 0xE1)
        self.assertEqual(encoded_array[3], 0xE0)
        self.assertEqual(encoded_array[4], 0x0)
        self.assertEqual(encoded_array[5], 0x0)
        self.assertEqual(encoded_array[6], 0x0)
        self.assertEqual(encoded_array[7], 0x0)
        self.assertEqual(encoded_array[8], 0x40)
        self.assertEqual(encoded_array[9], 0x6E)
        self.assertEqual(encoded_array[10], 0xBE)
        self.assertEqual(encoded_array[11], 0xAE)
        self.assertEqual(encoded_array[12], 0xFF)
        self.assertEqual(encoded_array[13], 0xFF)

    def test_decode_slof(self):
        test_array = [
            0x40,
            0xC3,
            0x88,
            0x0,
            0x0,
            0x0,
            0x0,
            0x0,
            0x13,
            0x1B,
            0xEA,
            0x2A,
            0xDE,
            0x3E,
        ]
        test_array = bytearray(test_array)
        self.Decoder.encoded_data = test_array
        decoded_array = self.Decoder.decode_slof()
        for i, dec in enumerate(decoded_array):
            self.assertAlmostEqual(dec, self.i_slof_data[i], places=2)
        # self.assertCountEqual(decoded_array, self.i_slof_data)

    def test_encode_decode_slof(self):
        """ """
        self.Decoder.decoded_data = self.i_slof_data
        self.Decoder.encode_slof()
        decoded_array = self.Decoder.decode_slof()
        for i, dec in enumerate(decoded_array):
            self.assertAlmostEqual(dec, self.i_slof_data[i], places=2)

    def test_encode_pic_i_data(self):
        self.Decoder.decoded_data = self.i_data
        encoded_array = self.Decoder.encode_pic()

        self.assertEqual(
            len(encoded_array), 14, msg="{}".format([hex(x) for x in encoded_array])
        )

        self.assertEqual(encoded_array[0], 0x20)
        self.assertEqual(encoded_array[1], 0x4B)
        self.assertEqual(encoded_array[2], 0x4C)
        self.assertEqual(encoded_array[3], 0x43)
        self.assertEqual(encoded_array[4], 0x1)
        self.assertEqual(encoded_array[5], 0xBE)
        self.assertEqual(encoded_array[6], 0x92)
        self.assertEqual(encoded_array[7], 0x8)
        self.assertEqual(encoded_array[8], 0x48)
        self.assertEqual(encoded_array[9], 0xE1)
        self.assertEqual(encoded_array[10], 0x20)
        self.assertEqual(encoded_array[11], 0xB)
        self.assertEqual(encoded_array[12], 0x17)
        self.assertEqual(encoded_array[13], 0xB0)

    def test_decode_pic_i_data(self):
        test_array = [
            0x20,
            0x4B,
            0x4C,
            0x43,
            0x1,
            0xBE,
            0x92,
            0x8,
            0x48,
            0xE1,
            0x20,
            0xB,
            0x17,
            0xB0,
        ]
        test_array = bytearray(test_array)
        self.Decoder.encoded_data = test_array
        decoded_array = self.Decoder.decode_pic()
        self.assertCountEqual(
            decoded_array,
            self.i_data,
            msg="{}\n{}".format([x for x in decoded_array], [x for x in self.i_data]),
        )

    def test_encode_decode_pic(self):
        """ """
        self.Decoder.decoded_data = self.i_slof_data
        self.Decoder.encode_pic()
        decoded_array = self.Decoder.decode_pic()
        self.assertCountEqual(self.i_slof_data, decoded_array)

    def test_encode_decode_fixed_point(self):
        encoded_fixed_point = self.Decoder._encode_fixed_point(self.fixed_point)
        decoded_fixed_point = self.Decoder._decode_fixed_point(encoded_fixed_point)
        self.assertEqual(self.fixed_point, decoded_fixed_point)

    # def test_encode_linear(self):
    #     self.Decoder.decoded_data = self.mz_data
    #     self.Decoder.encoded_data = None
    #     encoded_array = self.Decoder.encode_linear()
    #     self.assertEqual(
    #         len(encoded_array),
    #         23,
    #         msg='{0}\n{1}'.format(
    #             [hex(x) for x in encoded_array],
    #             [
    #                 '0x41', '0x74', '0x75', '0xa4', '0x70', '0x0', '0x0',
    #                 '0x0', '0xf6', '0xff', '0xff', '0x7f', '0xc2', '0x89',
    #                 '0xe2', '0x7f', '0x2b', '0xf3', '0x8a', '0x13', '0xdc',
    #                 '0x6a', '0x20'
    #             ]
    #         )
    #     )

    #     # assert first value
    #     self.assertEqual(
    #         0xff & encoded_array[8],
    #         0xf6,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             8, hex(0xff & encoded_array[8]), hex(0xf6)
    #         )
    #     )

    #     self.assertEqual(
    #         0xff & encoded_array[9],
    #         0xff,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             9,
    #             hex(0xff & encoded_array[9]),
    #             hex(0xff)
    #         )
    #         )
    #     self.assertEqual(
    #         0xff & encoded_array[10],
    #         0xff,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             10,
    #             hex(0xff & encoded_array[10]),
    #             hex(0xff)
    #         )
    #         )
    #     self.assertEqual(
    #         0xff & encoded_array[11],
    #         0x7f,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             11,
    #             hex(0xff & encoded_array[11]),
    #             hex(0x7f)
    #         )
    #         )

    #     # assert second value
    #     self.assertEqual(
    #         0xff & encoded_array[12],
    #         0xc2,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             12,
    #             hex(0xff & encoded_array[12]),
    #             hex(0xc2)
    #         )
    #     )
    #     self.assertEqual(
    #         0xff & encoded_array[13],
    #         0x89,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             13,
    #             hex(0xff & encoded_array[13]),
    #             hex(0x89)
    #         )
    #     )
    #     self.assertEqual(
    #         0xff & encoded_array[14],
    #         0xe2,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             14,
    #             hex(0xff & encoded_array[14]),
    #             hex(0xe2)
    #         )
    #     )
    #     self.assertEqual(
    #         0xff & encoded_array[15],
    #         0x7f,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             15,
    #             hex(0xff & encoded_array[15]),
    #             hex(0x7f)
    #         )
    #     )

    #     # assert third value
    #     self.assertEqual(
    #         0xff & encoded_array[16],
    #         0x2b,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             16,
    #             hex(0xff & encoded_array[16]),
    #             hex(0x2b)
    #         )
    #     )
    #     self.assertEqual(
    #         0xff & encoded_array[17],
    #         0xf3,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             17,
    #             hex(0xff & encoded_array[17]),
    #             hex(0xc2)
    #         )
    #     )
    #     self.assertEqual(
    #         0xff & encoded_array[18],
    #         0x8a,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             18,
    #             hex(0xff & encoded_array[18]),
    #             hex(0x8a)
    #         )
    #     )
    #     self.assertEqual(
    #         0xff & encoded_array[19],
    #         0x13,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             19,
    #             hex(0xff & encoded_array[19]),
    #             hex(0x13)
    #         )
    #     )

    #     # assert fourth value
    #     self.assertEqual(
    #         0xff & encoded_array[20],
    #         0xdc,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             19,
    #             hex(0xff & encoded_array[20]),
    #             hex(0xdc)
    #         )
    #     )
    #     self.assertEqual(
    #         0xff & encoded_array[21],
    #         0x6a,
    #         msg='Fail in value at pos {0}: {1} != {2}'.format(
    #             19,
    #             hex(0xff & encoded_array[21]),
    #             hex(0x6a)
    #         )
    #     )
    #     # self.assertEqual(
    #     #     0xff & encoded_array[22],
    #     #     0x20,
    #     #     msg='Fail in value at pos {0}: {1} != {2}'.format(
    #     #         19,
    #     #         hex(0xff & encoded_array[22]),
    #     #         hex(0x20)
    #     #     )
    #     # )

    # def test_decode_linear(self):

    #     # ouput of  >>> PyMSNumpress.encode_linear(self.mz_data, enc,
    #     # self.fixed_point)
    #     encoded_array = [
    #         65, 116, 117, 164, 112, 0, 0, 0, 246, 255,
    #         255, 127, 194, 137, 226, 127, 43, 243, 138, 19, 220, 106, 32
    #     ]

    #     # make bytearray from hex vals
    #     encoded_array = bytearray(encoded_array)
    #     MSNumpress.encoded_data = encoded_array

    #     # decode bytearray to numpy array
    #     decoded_array = self.Decoder.decode_linear()
    #     self.assertIsInstance(decoded_array, np.ndarray)
    #     self.assertEqual(len(decoded_array), len(self.mz_data))
    #     for i in range(len(decoded_array)):
    #         self.assertAlmostEqual(
    #             decoded_array[i], self.mz_data[i], places=i+1)

    def test_encode_decode_linear(self):
        """ """
        test_array = [
            100.00066,
            100.00217,
            100.00368,
            100.00519,
            111.73335,
            111.73513,
            111.73692,
            111.7387,
            111.74049,
            111.74227,
            111.74406,
            111.74584,
            111.74763,
            111.74941,
            111.7512,
            111.75298,
            111.75477,
            112.00694,
            112.00873,
            112.01052,
            112.01231,
            112.0141,
            112.01589,
            112.01768,
        ]
        self.Decoder.decoded_data = test_array
        self.Decoder.encode_linear()
        decoded_array = self.Decoder.decode_linear()
        for i in range(len(decoded_array)):
            self.assertAlmostEqual(
                decoded_array[i],
                test_array[i],
                places=4,
                msg="error at pos {0}".format(i),
            )

    def test_encode_decode_self_mz(self):
        self.Decoder.decoded_data = self.mz_data
        self.Decoder.encode_linear()
        decoded_array = self.Decoder.decode_linear()
        for i in range(len(decoded_array)):
            self.assertAlmostEqual(
                decoded_array[i],
                self.mz_data[i],
                places=4,
                msg="error at pos {0}".format(i),
            )


if __name__ == "__main__":
    unittest.main(verbosity=3)



# -*- coding: utf-8 -*-
# encoding: utf-8
"""
Class to parse the obo file and set up the accessions library

The OBO parser has been designed to convert MS:xxxxx tags to their appropriate
names. A minimal set of MS accession is used in pymzML, but additional
accessions can easily be queried.

The obo translator is used internally to associate names with MS:xxxxxxx tags.

The OboTranslator Class generates a dictionary and several lookup tables.
e.g.

::

    >>> from pymzml.obo import OboTranslator as OT
    >>> translator = OT()
    >>> translator['MS:1000127']
    'centroid mass spectrum'
    >>> translator['positive scan']
    {'is_a': 'MS:1000465 ! scan polarity', 'id': 'MS:1000130', 'def':
    '"Polarity of the scan is positive." [PSI:MS]', 'name': 'positive scan'}
    >>> translator['scan']
    {'relationship': 'part_of MS:0000000 ! Proteomics Standards Initiative Mass
    Spectrometry Ontology', 'id': 'MS:1000441', 'def': '"Function or process of
    the mass spectrometer where it records a spectrum." [PSI:MS]', 'name':
    'scan'}
    >>> translator['unit']
    {'relationship': 'part_of MS:0000000 ! Proteomics Standards Initiative Mass
    Spectrometry Ontology', 'id': 'MS:1000460', 'def': '"Terms to describe
    units." [PSI:MS]', 'name': 'unit'}

pymzML comes with the queryOBO.py script that can be used to interrogate the
OBO file. Please refer to :ref:`example_scripts` for further usage information.

::

    $ ./example_scripts/queryOBO.py "scan time"
    MS:1000016
    scan time
    "The time taken for an acquisition by scanning analyzers." [PSI:MS]
    Is a: MS:1000503 ! scan attribute
    $


::

    $ ./example_scripts/queryOBO.py 1000016
    MS:1000016
    scan time
    "The time taken for an acquisition by scanning analyzers." [PSI:MS]
    MS:1000503 ! scan attribute
    $


"""


# Python mzML module - pymzml
# Copyright (C) 2010-2019 M. Kösters, C. Fufezan
#     The MIT License (MIT)

#     Permission is hereby granted, free of charge, to any person obtaining a copy
#     of this software and associated documentation files (the "Software"), to deal
#     in the Software without restriction, including without limitation the rights
#     to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#     copies of the Software, and to permit persons to whom the Software is
#     furnished to do so, subject to the following conditions:

#     The above copyright notice and this permission notice shall be included in all
#     copies or substantial portions of the Software.

#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#     IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#     FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#     AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#     LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#     OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#     SOFTWARE.


import sys
import os
import re
import gzip
import urllib


class OboTranslator(object):
    """
    Generates a mapping from MS:xxxxx to names and vice versa for a specific
    obo version

    Args:
        version (str): obo version
    """

    _obo_instance_cache = {}

    def __init__(self, version=None):
        self.version = self.__normalize_version(version)
        self.all_dicts = []
        self.id = {}
        self.name = {}
        self.definition = {}
        self.lookups = [self.id, self.name, self.definition]
        self.MS_tag_regex = re.compile(r"MS:[0-9]*")

        # Only parse the OBO when necessary, not upon object construction
        self.__obo_parsed = False

    @classmethod
    def from_cache(cls, version):
        version = cls.__normalize_version(version)
        try:
            return cls._obo_instance_cache[version]
        except KeyError:
            inst = cls._obo_instance_cache[version] = cls(version)
            return inst

    def __setitem__(self, key, value):
        raise TypeError("OBO translator dictionaries only support assignment via .add")

    def __getitem__(self, key):
        if not self.__obo_parsed:
            self.parseOBO()

        for lookup in self.lookups:
            if key in lookup:
                if self.MS_tag_regex.match(key):
                    try:
                        return lookup[key]["name"]
                    except:
                        pass
                return lookup[key]
        return None

    @staticmethod
    def __normalize_version(version):
        """
        Ensure that a version has 3 components, defaulting to .0 for the
        missing components.

        Args:
            version (str): The original version to modify.

        Returns:
            version (str): The version, normalized to ensure that it has 3
                parts.
        """
        if version is not None:
            parts = version.split(".")

            missing_parts = 3 - len(parts)
            if missing_parts > 0:
                version = ".".join(parts + ["0"] * missing_parts)

        return version

    def download_obo(self, version, obo_file):
        uri = f"https://raw.githubusercontent.com/pymzml/psi-ms-CV/v{self.version}/psi-ms.obo"
        urllib.request.urlretrieve(uri, obo_file)

        with open(obo_file, "rb") as fin, gzip.open(obo_file + ".gz", "wb") as fout:
            fout.writelines(fin.readlines())
            os.remove(obo_file)
        return

    def parseOBO(self):
        self.__obo_parsed = True
        """
        Locate and parse the OBO file in the OBO root directory.

        Note:

           cx_Freeze friendly. If using cx_Freeze, place the OBO folder at
           the location of sys.executable.
        """

        # TODO: Try to get all the versions, even those without well-defined
        #       version numbers, or get remote hosting of all of the versions
        #       and only download one at will on demand.

        # Modify the root for cx_freeze
        if getattr(sys, "frozen", False):
            obo_root = os.path.dirname(sys.executable)
        else:
            obo_root = os.path.dirname(__file__)

        obo_file = os.path.join(
            obo_root,
            "obo",
            "psi-ms{0}.obo".format("-" + self.version if self.version else ""),
        )
        if os.path.exists(obo_file):
            pass
        elif os.path.exists(obo_file + ".gz"):
            obo_file = obo_file + ".gz"
        else:
            self.download_obo(self.version, obo_file)
            obo_file += ".gz"

        with open(obo_file, "rb") as fin:
            # never rely on file extensions!
            first_two_bytes = fin.read(2)
            # check if file is gzipped by magic bytes
            if first_two_bytes == b"\x1f\x8b":
                open_func = gzip.open
            else:
                raise Exception(
                    "Filename has .gz extension but is missing the gzip magic bytes.\n"
                    "The file may be corrupted or not gzipped."
                )

        with open_func(obo_file, "rt", encoding="utf-8") as obo:
            collections = {}
            collect = False
            for line in obo:
                if line.strip() in ("[Term]", ""):
                    collect = True
                    if not collections:
                        continue
                    self.add(collections)
                    collections = {}
                else:
                    if line.strip() != "" and collect is True:
                        k = line.find(":")
                        collections[line[:k]] = line[k + 1 :].strip()
        return

    def add(self, collection_dict):
        """
        Add a new dict to the translator.

        Args:
            collection_dict (dict): python dict containing MS:xxxxxxx name
                mapping.
        """
        if not self.__obo_parsed:
            self.parseOBO()

        self.all_dicts.append(collection_dict)
        if "id" in collection_dict.keys():
            self.id[collection_dict["id"]] = self.all_dicts[-1]
        if "name" in collection_dict.keys():
            self.name[collection_dict["name"]] = self.all_dicts[-1]
        if "def" in collection_dict.keys():
            self.definition[collection_dict["def"]] = self.all_dicts[-1]

        return

    def checkOBO(self, idTag, name):
        """
        Check if idTag equals name in currently used obo version.

        Args:
            idTag (str): MS accession tag
            name  (str): trivial name of the accession

        Returns:
            boolean: True if idTag and name correspond, else False.
        """
        if not self.__obo_parsed:
            self.parseOBO()

        if self.id[idTag]["name"] == name:
            return True
        else:
            return False


if __name__ == "__main__":
    print(__doc__)
